import strax
import numpy as np
from immutabledict import immutabledict
import straxen

export, __all__ = strax.exporter()


@export
class OnlineMonitorNV(strax.Plugin):
    """
    Plugin to write data of nVeto detector to the online-monitor. 
    Data that is written by this plugin should be small (~MB/chunk) 
    to not overload the runs-database.

    This plugin takes 'hitlets_nv' and 'events_nv'. Although they are
    not strictly related, they are aggregated into a single data_type
    in order to minimize the number of documents in the online monitor.

    Produces 'online_monitor_nv' with info on the hitlets_nv and events_nv
    """

    __version__ = '0.0.4'

    depends_on = ('hitlets_nv', 'events_nv')
    provides = 'online_monitor_nv'
    data_kind = 'online_monitor_nv'

    # Needed in case we make again an muVETO child.
    ends_with = '_nv'

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help='immutabledict mapping subdetector to (min, max) '
             'channel number.')

    events_area_bounds = straxen.URLConfig(
        type=tuple, default=(-0.5, 130.5),
        help='Boundaries area histogram of events_nv_area_per_chunk [PE]')

    events_area_nbins = straxen.URLConfig(
        type=int, default=131,
        help='Number of bins of histogram of events_nv_area_per_chunk, '
             'defined value 1 PE/bin')

    def infer_dtype(self):
        self.channel_range = self.channel_map['nveto']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_monitor_dtype(self.ends_with, self.n_channel, self.events_area_nbins)

    def compute(self, hitlets_nv, events_nv, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res['time'] = start
        res['endtime'] = end

        # Count number of hitlets_nv per PMT
        hitlets_channel_count, _ = np.histogram(hitlets_nv['channel'],
                                                bins=self.n_channel,
                                                range=[self.channel_range[0],
                                                       self.channel_range[1] + 1])
        res[f'hitlets{self.ends_with}_per_channel'] = hitlets_channel_count

        # Count number of events_nv with coincidence cut
        res[f'events{self.ends_with}_per_chunk'] = len(events_nv)
        sel = events_nv['n_contributing_pmt'] >= 4
        res[f'events{self.ends_with}_4coinc_per_chunk'] = np.sum(sel)
        sel = events_nv['n_contributing_pmt'] >= 5
        res[f'events{self.ends_with}_5coinc_per_chunk'] = np.sum(sel)
        sel = events_nv['n_contributing_pmt'] >= 8
        res[f'events{self.ends_with}_8coinc_per_chunk'] = np.sum(sel)
        sel = events_nv['n_contributing_pmt'] >= 10
        res[f'events{self.ends_with}_10coinc_per_chunk'] = np.sum(sel)

        # Get histogram of events_nv_area per chunk
        events_area, bins_ = np.histogram(events_nv['area'],
                                          bins=self.events_area_nbins,
                                          range=self.events_area_bounds)
        res[f'events{self.ends_with}_area_per_chunk'] = events_area
        return res


def veto_monitor_dtype(veto_name: str = '_nv',
                       n_pmts: int = 120,
                       n_bins: int = 131) -> list:
    dtype = []
    dtype += strax.time_fields  # because mutable
    dtype += [((f'hitlets{veto_name} per channel', f'hitlets{veto_name}_per_channel'), (np.int64, n_pmts)),
              ((f'events{veto_name}_area per chunk', f'events{veto_name}_area_per_chunk'), np.int64, n_bins),
              ((f'events{veto_name} per chunk', f'events{veto_name}_per_chunk'), np.int64),
              ((f'events{veto_name} 4-coincidence per chunk', f'events{veto_name}_4coinc_per_chunk'), np.int64),
              ((f'events{veto_name} 5-coincidence per chunk', f'events{veto_name}_5coinc_per_chunk'), np.int64),
              ((f'events{veto_name} 8-coincidence per chunk', f'events{veto_name}_8coinc_per_chunk'), np.int64),
              ((f'events{veto_name} 10-coincidence per chunk', f'events{veto_name}_10coinc_per_chunk'), np.int64)
              ]
    return dtype



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class EnergyEstimates(strax.Plugin):
    """
    Plugin which converts cS1 and cS2 into energies (from PE to KeVee).
    """
    __version__ = '0.1.1'
    depends_on = ['corrected_areas']
    dtype = [
                ('e_light', np.float32, 'Energy in light signal [keVee]'),
                ('e_charge', np.float32, 'Energy in charge signal [keVee]'),
                ('e_ces', np.float32, 'Energy estimate [keVee]')
            ] + strax.time_fields
    save_when = strax.SaveWhen.TARGET

    # config options don't double cache things from the resource cache!
    g1 = straxen.URLConfig(
        default='bodega://g1?bodega_version=v2',
        help="S1 gain in PE / photons produced",
    )
    g2 = straxen.URLConfig(
        default='bodega://g2?bodega_version=v2',
        help="S2 gain in PE / electrons produced",
    )
    lxe_w = straxen.URLConfig(
        default=13.7e-3,
        help="LXe work function in quanta/keV"
    )

    def compute(self, events):
        el = self.cs1_to_e(events['cs1'])
        ec = self.cs2_to_e(events['cs2'])
        return dict(e_light=el,
                    e_charge=ec,
                    e_ces=el + ec,
                    time=events['time'],
                    endtime=strax.endtime(events))

    def cs1_to_e(self, x):
        return self.lxe_w * x / self.g1

    def cs2_to_e(self, x):
        return self.lxe_w * x / self.g2



"""Dynamic holoviews-based waveform display
Note imports are inside function, to keep 'import straxen'
free of holoviews.
"""

import numpy as np
import pandas as pd

import straxen

straxen._BOKEH_X_RANGE = None


@straxen.mini_analysis(requires=['hitlets_nv', 'events_nv', 'event_positions_nv'])
def plot_nveto_event_display(context, run_id, hitlets_nv, events_nv):
    nvd = straxen.nVETOEventDisplay(events_nv, hitlets_nv, run_id)
    return nvd.plot_event_display()


def seconds_from(t, t_reference, unit_conversion=int(1e9)):
    """
    Helper function which concerts times into relative times in
    specified unit.

    :param t: Time
    :param t_reference: Reference time e.g. start of an event or first
        peak in event.
    :param unit_conversion: Conversion factor for time units e.g. 10**3
        for micro seconds.
    """
    return (t - t_reference) / unit_conversion


# Custom wheel zoom tool that only zooms in one dimension
def x_zoom_wheel():
    import bokeh.models
    return bokeh.models.WheelZoomTool(dimensions='width')


@straxen.mini_analysis(requires=['records'], hv_bokeh=True)
def hvdisp_plot_pmt_pattern(*, config, records, to_pe, array='bottom'):
    """Plot a PMT array, with colors showing the intensity
    of light observed in the time range
    :param array: 'top' or 'bottom', array to show
    """
    import holoviews as hv

    pmts = straxen.pmt_positions(xenon1t=config['n_tpc_pmts'] < 300)
    areas = np.bincount(records['channel'],
                        weights=records['area'] * to_pe[records['channel']],
                        minlength=len(pmts))

    # Which PMTs should we include?
    m = pmts['array'] == array
    pmts = pmts[m].copy()
    pmts['area'] = areas[m]

    f = 1.08
    pmts = hv.Dataset(
        pmts,
        kdims=[hv.Dimension('x',
                            unit='cm',
                            range=(-straxen.tpc_r * f, straxen.tpc_r * f)),
               hv.Dimension('y',
                            unit='cm',
                            range=(-straxen.tpc_r * f, straxen.tpc_r * f)),
               hv.Dimension('i', range=(0, config['n_tpc_pmts']), label='PMT number'),
               hv.Dimension('area', label='Area', unit='PE')])
    pmts = pmts.to(
        hv.Points,
        vdims=['area', 'i'],
        group='PMTPattern',
        label=array.capitalize()).opts(
        plot=dict(color_index=2,
                  tools=['hover'],
                  show_grid=False),
        style=dict(size=17,
                   cmap='plasma'))

    return pmts


def _records_to_points(*, records, to_pe, t_reference,
                       config, unit_conversion=int(1e9)):
    """Return (holoviews.Points, time_stream) corresponding to records.
    """
    import holoviews as hv

    areas_r = records['area'] * to_pe[records['channel']]

    # Create dataframe with record metadata
    df = pd.DataFrame(dict(
        area=areas_r,
        time=seconds_from(records['time'] + records['dt'] * records['length'] // 2,
                          t_reference, unit_conversion=unit_conversion),
        length=records['length'],
        dt=records['dt'],
        channel=records['channel']))

    rec_points = hv.Points(
        df,
        kdims=[hv.Dimension('time', label='Time [µs]'),
               hv.Dimension('channel',
                            label='PMT Channel',
                            range=(0, config['n_tpc_pmts']))],
        vdims=[hv.Dimension('area', label='Area [pe]')])

    time_stream = hv.streams.RangeX(source=rec_points)
    return rec_points, time_stream


@straxen.mini_analysis(requires=['records'], hv_bokeh=True)
def hvdisp_plot_records_2d(records,
                           to_pe,
                           config,
                           t_reference,
                           time_stream=None,
                           tools=(x_zoom_wheel(), 'xpan'),
                           default_tools=('save', 'pan', 'box_zoom', 'save', 'reset'),
                           plot_library='bokeh',
                           hooks=()):
    """Plot records in a dynamic 2D histogram of (time, pmt)
    :param width: Plot width in pixels
    :param time_stream: holoviews rangex stream to use. If provided,
        we assume records is already converted to points (which hopefully
        is what the stream is derived from)
    :param tools: Tools to be used in the interactive plot. Only works
        with bokeh as plot library.
    :param plot_library: Default bokeh, library to be used for the
        plotting.
    :param width: With of the record matrix in pixel.
    :param hooks: Hooks to adjust plot settings.
    :returns: datashader object, records holoview points,
        RangeX time stream of records.
    """
    shader, records, time_stream = _hvdisp_plot_records_2d(records,
                                                           to_pe,
                                                           config,
                                                           t_reference,
                                                           time_stream=time_stream,
                                                           default_tools=default_tools,
                                                           tools=tools,
                                                           hooks=hooks,
                                                           plot_library=plot_library)
    shader = shader.opts(title="Time vs. Channel")
    return shader


def _hvdisp_plot_records_2d(records,
                            to_pe,
                            config,
                            t_reference,
                            event_range=(None, None),
                            time_stream=None,
                            default_tools=(),
                            tools=(),
                            hooks=(),
                            plot_library='bokeh',
                            unit_conversion=10**3):
    import holoviews as hv
    import holoviews.operation.datashader
    hv.extension(plot_library)

    if time_stream is None:
        # Records are still a dataframe, convert it to points
        records, time_stream = _records_to_points(records=records,
                                                  to_pe=to_pe,
                                                  t_reference=t_reference,
                                                  unit_conversion=unit_conversion,
                                                  config=config)

    # Whether to show the toolbar or not:
    if tools:
        toolbar = 'right'
    else:
        toolbar = None

    # Creating the plot:
    shader = hv.operation.datashader.dynspread(
        hv.operation.datashader.datashade(
            records,
            dynamic=True,
            x_range=event_range,
            y_range=(0, config['n_tpc_pmts']),
            streams=[time_stream]), threshold=0.1).opts(
        plot=dict(aspect=4,
                  responsive='width',
                  hooks=list(hooks),
                  toolbar=toolbar,
                  tools=list(tools),
                  default_tools=list(default_tools),
                  fontsize={'labels': 12},
                  show_grid=True))

    return shader, records, time_stream


def plot_record_polygons(record_points,
                         center_time=True,
                         scaling=10**-3,
                         ):
    """
    Plots record hv.Points as polygons for record matrix.

    :param record_points: Holoviews Points generated with
        _records_to_points.
    :param center_time: If true treats specified times as center times.
    :param scaling: Scale times scale by this factor. E.g. if times
        should be in µs use 10**-3.
    :returns: hv.Polygons
    """
    import holoviews as hv

    df = record_points.data.copy()
    if not center_time:
        df.loc[:, 'time'] = df.loc[:, 'time'] + (df.loc[:, 'length'] / 2 * df.loc[:, 'dt'])

    data = _make_polys(df,
                       scaling=scaling)
    polys = hv.Polygons(data,
                        kdims=['time', 'channel'],
                        vdims=list(data[0].keys())[1:]).opts(color='area',
                                                             aspect=4,
                                                             responsive='width',
                                                             line_width=0,
                                                             cmap='viridis',
                                                             )

    return polys


def _make_polys(df, scaling=1):
    """
    Function which converts the hitlet/record points into polygons
    for the hitlet/record matrix.
    """
    res = []
    for row_i in range(len(df)):
        data_dict = {}
        # Kdims are given by time and channel:
        t, ch = df.loc[row_i, ['time', 'channel']]
        w = df.loc[row_i, 'length'] * df.loc[row_i, 'dt']
        data_dict[('time', 'channel')] = _rectangle(t, ch, width=w * scaling)

        # Add other parameters to polygons for vdmis:
        keys = [key for key in df.columns.values if key not in ['time',
                                                                'channel',
                                                                'dt']]
        for key in keys:
            data_dict[key] = df.loc[row_i, key]
        res.append(data_dict)
    return res


def _rectangle(time=0, channel=0, width=1.1, height=1):
    """
    Generates polygon box coordinates for record matrix.

    :param time: Center position of the record in time.
    :param channel: Center position of PMT channel. E.g channel 0 => 0.5
    :param width: Length of the record in µs.
    param height: Height of the record in "channel"-units.

    X,Y have to be the center of the polygon.
    Width and height are the full width and height in data coordinates.
    """
    width = width / 2
    height = height / 2
    return np.array([(time - width, channel - height),
                     (time + width, channel - height),
                     (time + width, channel + height),
                     (time - width, channel + height)])


def get_records_matrix_in_window(polys, x_range, time_slice=10):
    """
    Helper function which returns polygons for rendering when x_range
    is below the specified value.

    This function has to be applied to polygons e.g.:

        poly.apply(get_records_matrix_in_window, streams=[time_stream])

    :param polys: Holoviews Polygons
    :param x_range: x_range of the RangeX object.
    :param time_slice: Size of the time slice in [µs] when records
        should be drawn.
    """
    if x_range is None:
        # Needed since x_range is by default not defined when plotting
        # the first time.
        return polys.iloc[:0]
    if (x_range[1] - x_range[0]) < time_slice:
        # If x_range smaller than specified minimum return polys ->
        # render polys.
        inds = _in_window(polys.data, x_range)
        return polys.iloc[inds]
    return polys.iloc[:0]


def _in_window(polys, x_range):
    """
    Function which checks if a polygon is partially in x_range.

    :param polys: List of ordered dictionaries containing Polygon data.
    :param x_range: Range which should be tested.
    """
    res = []
    for ind, p in enumerate(polys):
        # Loop over polys, if partially in window keep index.
        if np.any((x_range[0] <= p['time']) & (p['time'] < x_range[1])):
            res.append(ind)
    return res


@straxen.mini_analysis(
    requires=['peaks', 'peak_basics'],
    hv_bokeh=True)
def hvdisp_plot_peak_waveforms(
        t_reference,
        time_range,
        peaks,
        width=600,
        show_largest=None,
        time_dim=None):
    """Plot the sum waveforms of peaks
    :param width: Plot width in pixels
    :param show_largest: Maximum number of peaks to show
    :param time_dim: Holoviews time dimension; will create new one
    if not provided.
    """
    import holoviews as hv

    if show_largest is not None and len(peaks) > show_largest:
        show_i = np.argsort(peaks['area'])[-show_largest::]
        peaks = peaks[show_i]

    curves = []
    for p in peaks:
        # label = {1: 's1', 2: 's2'}.get(
        #     p['type'], 'unknown')
        color = {1: 'b', 2: 'g'}.get(
            p['type'], 'k')

        # It's better to plot amplitude /time than per bin, since
        # sampling times are now variable
        y = p['data'][:p['length']] / p['dt']
        t_edges = np.arange(p['length'] + 1, dtype=np.int64)
        t_edges = t_edges * p['dt'] + p['time']
        t_edges = seconds_from(t_edges, t_reference)

        # Make a 'step' plot. Unlike matplotlib's steps-mid,
        # this also analyses the final edges correctly
        t_ = np.zeros(2 * len(y))
        y_ = np.zeros(2 * len(y))
        t_[0::2] = t_edges[:-1]
        t_[1::2] = t_edges[1:]
        y_[0::2] = y
        y_[1::2] = y

        if time_dim is None:
            time_dim = hv.Dimension('time', label='Time', unit='sec')

        curves.append(
            hv.Curve(dict(time=t_, amplitude=y_),
                     kdims=time_dim,
                     vdims=hv.Dimension('amplitude', label='Amplitude',
                                        unit='PE/ns'),
                     group='PeakSumWaveform').opts(style=dict(color=color)))

    return hv.Overlay(items=curves).opts(plot=dict(width=width))


def _range_plot(f, full_time_range, t_reference, **kwargs):
    # The **bla is needed to disable some holoviews check
    # on the arguments...
    def wrapped(x_range, **kwargzz):
        if len(kwargzz):
            raise RuntimeError(f"Passed superfluous kwargs {kwargzz}")
        if x_range is None:
            x_range = seconds_from(np.asarray(full_time_range),
                                   t_reference)

        # Deal with strange time ranges -- not sure how these arise?
        x_range = np.nan_to_num(x_range)
        if x_range[1] == x_range[0]:
            x_range[1] += 1

        return f(time_range=(t_reference + int(x_range[0] * 1e9),
                             t_reference + int(x_range[1] * 1e9)),
                 t_reference=t_reference,
                 **kwargs)

    return wrapped


@straxen.mini_analysis(
    requires=['records', 'peaks', 'peak_basics'],
    hv_bokeh=True)
def waveform_display(
        context, run_id, to_pe, time_range, t_reference, records, peaks,
        config,
        width=600, show_largest=None):
    """Plot a waveform overview display"
    :param width: Plot width in pixels
    """
    import holoviews as hv

    records_points, time_stream = _records_to_points(records=records,
                                                     to_pe=to_pe,
                                                     t_reference=t_reference,
                                                     config=config)

    time_v_channel = context.hvdisp_plot_records_2d(
        run_id=run_id, to_pe=to_pe,
        records=records_points,
        time_stream=time_stream,
        time_range=time_range, t_reference=t_reference,
        # We don't need to cut these further, records we get are already cut
        # to the plot's maximum range by the mini_analysis logic
        # and datashader does the online cutting / rebinning / zooming.
        # This is fortunate, since we omitted 'endtime' from records_points!
        time_selection='skip')

    array_plot = {
        array: hv.DynamicMap(
            _range_plot(
                context.hvdisp_plot_pmt_pattern,
                run_id=run_id, to_pe=to_pe,
                records=records,
                full_time_range=time_range,
                t_reference=t_reference,
                time_selection='touching',
                array=array),
            streams=[time_stream])
        for array in ('top', 'bottom')}

    peak_wvs = hv.DynamicMap(
        _range_plot(
            context.hvdisp_plot_peak_waveforms,
            run_id=run_id,
            width=width,
            full_time_range=time_range,
            t_reference=t_reference,
            time_selection='touching',
            time_dim=records_points.kdims[0],
            peaks=peaks,
            show_largest=show_largest),
        streams=[time_stream])

    layout = time_v_channel + peak_wvs + array_plot['top'] + array_plot['bottom']
    return layout.cols(2)


def hook(plot, x_range, debug=False):
    """
    Hook to set the same RangeX stream for event display and
    records matrix, voodoo....

    Note:
        Works only in the following order:

        1. Create holoviews
        2. hv.render plot
        3. set bokeh x_range as holoviews x_range

        Does not work first with bokeh and then with holoviews. Why?
        I have no clue....
    """
    if debug:
        print('Hook: ', x_range)
    if not x_range:
        _hook_get_xrange(plot, debug)
    else:
        _hook_set_xrange(plot, x_range, debug)


def _hook_get_xrange(plot, debug):
    straxen._BOKEH_X_RANGE = plot.handles['x_range']
    if debug:
        print('Get', straxen._BOKEH_X_RANGE)


def _hook_set_xrange(plot, x_range, debug):
    if debug:
        print('Set', x_range)
    plot.state.x_range = x_range



import strax
import straxen

from straxen.plugins.defaults import MV_PREAMBLE, NV_HIT_DEFAULTS, MV_HIT_DEFAULTS
from straxen.plugins.records_nv.records_nv import nVETOPulseProcessing


export, __all__ = strax.exporter()


@export
class muVETOPulseProcessing(nVETOPulseProcessing):
    __doc__ = MV_PREAMBLE + nVETOPulseProcessing.__doc__
    __version__ = '0.0.1'

    depends_on = 'raw_records_mv'
    provides = 'records_mv'
    data_kind = 'records_mv'
    child_plugin = True

    save_outside_hits_mv = straxen.URLConfig(
        default=MV_HIT_DEFAULTS['save_outside_hits_mv'], track=True, infer_type=False,
        child_option=True, parent_option_name='save_outside_hits_nv',
        help='Save (left, right) samples besides hits; cut the rest')

    baseline_samples_mv = straxen.URLConfig(
        default=100, track=True, infer_type=False,
        child_option=True, parent_option_name='baseline_samples_nv',
        help='Number of samples to use at the start of the pulse to determine '
             'the baseline')

    hit_min_amplitude_mv = straxen.URLConfig(
        infer_type=False,
        default=MV_HIT_DEFAULTS['hit_min_amplitude_mv'], track=True,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_mveto_pmts, or a number, '
             'or a string like "pmt_commissioning_initial" which means calling '
             'hitfinder_thresholds.py, '
             'or a tuple like (correction=str, version=str, nT=boolean),'
             'which means we are using cmt.')

    check_raw_record_overlaps = straxen.URLConfig(
        default=True, track=False, infer_type=False,
        help='Crash if any of the pulses in raw_records overlap with others '
             'in the same channel')

    def setup(self):
        self.baseline_samples = self.baseline_samples_mv
        self.hit_thresholds = self.hit_min_amplitude_mv

    def infer_dtype(self):
        record_length = strax.record_length_from_dtype(
            self.deps['raw_records_mv'].dtype_for('raw_records_mv'))
        dtype = strax.record_dtype(record_length)
        return dtype

    def compute(self, raw_records_mv):
        if self.check_raw_record_overlaps:
            straxen.check_overlaps(raw_records_mv, n_channels=3000)
        return super().compute(raw_records_mv)



import numpy as np
import numba
import strax
import straxen


export, __all__ = strax.exporter()


@export
class PeakBasics(strax.Plugin):
    """
    Compute the basic peak-properties, thereby dropping structured
    arrays.
    NB: This plugin can therefore be loaded as a pandas DataFrame.
    """
    __version__ = '0.1.4'
    parallel = True
    depends_on = ('peaks',)
    provides = 'peak_basics'
    dtype = [
        (('Start time of the peak (ns since unix epoch)',
          'time'), np.int64),
        (('End time of the peak (ns since unix epoch)',
          'endtime'), np.int64),
        (('Weighted center time of the peak (ns since unix epoch)',
          'center_time'), np.int64),
        (('Peak integral in PE',
          'area'), np.float32),
        (('Number of hits contributing at least one sample to the peak',
          'n_hits'), np.int32),
        (('Number of PMTs contributing to the peak',
          'n_channels'), np.int16),
        (('PMT number which contributes the most PE',
          'max_pmt'), np.int16),
        (('Area of signal in the largest-contributing PMT (PE)',
          'max_pmt_area'), np.float32),
        (('Total number of saturated channels',
          'n_saturated_channels'), np.int16),
        (('Width (in ns) of the central 50% area of the peak',
          'range_50p_area'), np.float32),
        (('Width (in ns) of the central 90% area of the peak',
          'range_90p_area'), np.float32),
        (('Fraction of area seen by the top array '
          '(NaN for peaks with non-positive area)',
          'area_fraction_top'), np.float32),
        (('Length of the peak waveform in samples',
          'length'), np.int32),
        (('Time resolution of the peak waveform in ns',
          'dt'), np.int16),
        (('Time between 10% and 50% area quantiles [ns]',
          'rise_time'), np.float32),
        (('Number of PMTs with hits within tight range of mean',
          'tight_coincidence'), np.int16),
        (('Classification of the peak(let)',
          'type'), np.int8),
        (('Largest time difference between apexes of hits inside peak [ns]',
          'max_diff'), np.int32),
        (('Smallest time difference between apexes of hits inside peak [ns]',
          'min_diff'), np.int32),
    ]

    n_top_pmts = straxen.URLConfig(default=straxen.n_top_pmts, infer_type=False,
                                   help="Number of top PMTs")

    check_peak_sum_area_rtol = straxen.URLConfig(default=None, track=False, infer_type=False,
                                                 help="Check if the sum area and the sum of area per "
                                                      "channel are the same. If None, don't do the "
                                                      "check. To perform the check, set to the desired "
                                                      " rtol value used e.g. '1e-4' (see np.isclose).")

    def compute(self, peaks):
        p = peaks
        r = np.zeros(len(p), self.dtype)
        needed_fields = 'time length dt area type max_diff min_diff'
        for q in needed_fields.split():
            r[q] = p[q]
        r['endtime'] = p['time'] + p['dt'] * p['length']
        r['n_channels'] = (p['area_per_channel'] > 0).sum(axis=1)
        r['n_hits'] = p['n_hits']
        r['range_50p_area'] = p['width'][:, 5]
        r['range_90p_area'] = p['width'][:, 9]
        r['max_pmt'] = np.argmax(p['area_per_channel'], axis=1)
        r['max_pmt_area'] = np.max(p['area_per_channel'], axis=1)
        r['tight_coincidence'] = p['tight_coincidence']
        r['n_saturated_channels'] = p['n_saturated_channels']

        n_top = self.n_top_pmts
        area_top = p['area_per_channel'][:, :n_top].sum(axis=1)
        # Recalculate to prevent numerical inaccuracy #442
        area_total = p['area_per_channel'].sum(axis=1)
        # Negative-area peaks get NaN AFT
        m = p['area'] > 0
        r['area_fraction_top'][m] = area_top[m] / area_total[m]
        r['area_fraction_top'][~m] = float('nan')
        r['rise_time'] = -p['area_decile_from_midpoint'][:, 1]

        if self.check_peak_sum_area_rtol is not None:
            self.check_area(area_total, p, self.check_peak_sum_area_rtol)
        # Negative or zero-area peaks have centertime at startime
        r['center_time'] = p['time']
        r['center_time'][m] += self.compute_center_times(peaks[m])
        return r

    @staticmethod
    @numba.njit(cache=True, nogil=True)
    def compute_center_times(peaks):
        result = np.zeros(len(peaks), dtype=np.int32)
        for p_i, p in enumerate(peaks):
            t = 0
            for t_i, weight in enumerate(p['data']):
                t += t_i * p['dt'] * weight
            result[p_i] = t / p['area']
        return result

    @staticmethod
    def check_area(area_per_channel_sum, peaks, rtol) -> None:
        """
        Check if the area of the sum-wf is the same as the total area
            (if the area of the peak is positively defined).

        :param area_per_channel_sum: the summation of the
            peaks['area_per_channel'] which will be checked against the
             values of peaks['area'].
        :param peaks: array of peaks.
        :param rtol: relative tolerance for difference between
            area_per_channel_sum and peaks['area']. See np.isclose.
        :raises: ValueError if the peak area and the area-per-channel
            sum are not sufficiently close
        """
        positive_area = peaks['area'] > 0
        if not np.sum(positive_area):
            return

        is_close = np.isclose(area_per_channel_sum[positive_area],
                              peaks[positive_area]['area'],
                              rtol=rtol,
                              )

        if not is_close.all():
            for peak in peaks[positive_area][~is_close]:
                print('bad area')
                strax.print_record(peak)

            p_i = np.where(~is_close)[0][0]
            peak = peaks[positive_area][p_i]
            area_fraction_off = 1 - area_per_channel_sum[positive_area][p_i] / peak['area']
            message = (f'Area not calculated correctly, it\'s '
                       f'{100 * area_fraction_off} % off, time: {peak["time"]}')
            raise ValueError(message)



from straxen.plugins.events.corrected_areas import CorrectedAreas
import strax
import numpy as np
import straxen

export, __all__ = strax.exporter()


@export
class PeakCorrectedAreas(CorrectedAreas):
    """
    Pluging to apply corrections on peak level assuming that the main 
    S1 is the only physical S1. 
    """
    __version__ = '0.0.1'

    depends_on = ('peak_basics', 'peak_positions', 'peak_per_event')
    data_kind = 'peaks'
    provides = 'peak_corrections'

    electron_drift_velocity = straxen.URLConfig(
        default='cmt://'
                'electron_drift_velocity'
                '?version=ONLINE&run_id=plugin.run_id',
        cache=True,
        help='Vertical electron drift velocity in cm/ns (1e4 m/ms)'
    )

    electron_drift_time_gate = straxen.URLConfig(
        default='cmt://'
                'electron_drift_time_gate'
                '?version=ONLINE&run_id=plugin.run_id',
        help='Electron drift time from the gate in ns',
        cache=True)

    def infer_dtype(self):
        dtype = strax.time_fields + [
            (('Corrected area of S2 before elife correction '
              '(s2 xy correction + SEG/EE correction applied) [PE]',
              'cs2_wo_elifecorr'), np.float32),
            (('Corrected area of S2 before SEG/EE and elife corrections '
              '(s2 xy correction applied) [PE]',
              'cs2_wo_timecorr'), np.float32),
            (('Fraction of area seen by the top PMT array for corrected S2',
              'cs2_area_fraction_top'), np.float32),
            (('Corrected area of S2 in the bottom PMT array [PE]',
              'cs2_bottom'), np.float32),
            (('Corrected area of S2 [PE]', 'cs2'), np.float32),
            (('Correction factor for the S1 area based on S2 position',
              's1_xyz_correction_factor'), np.float32),
            (('Relative light yield correction factor for the S1 area',
              's1_rel_light_yield_correction_factor'), np.float32),
            (('z position of the multiscatter peak',
              'z_obs_ms'), np.float32),
        ]
        return dtype

    def compute(self, peaks):
        result = np.zeros(len(peaks), self.dtype)
        result['time'] = peaks['time']
        result['endtime'] = peaks['endtime']

        # Get z position of the peak
        z_obs = -self.electron_drift_velocity * peaks['drift_time']
        z_obs = z_obs + self.electron_drift_velocity * self.electron_drift_time_gate
        result['z_obs_ms'] = z_obs

        # Get S1 correction factors
        peak_positions = np.vstack([peaks['x'], peaks['y'], z_obs]).T
        result['s1_xyz_correction_factor'] = 1 / self.s1_xyz_map(peak_positions)
        result['s1_rel_light_yield_correction_factor'] = 1 / self.rel_light_yield

        # s2 corrections
        s2_top_map_name, s2_bottom_map_name = self.s2_map_names()

        seg, avg_seg, ee = self.seg_ee_correction_preparation()

        # now can start doing corrections

        # S2(x,y) corrections use the observed S2 positions
        s2_positions = np.vstack([peaks['x'], peaks['y']]).T

        # corrected s2 with s2 xy map only, i.e. no elife correction
        # this is for s2-only events which don't have drift time info

        cs2_top_xycorr = (peaks['area']
                          * peaks['area_fraction_top']
                          / self.s2_xy_map(s2_positions, map_name=s2_top_map_name))
        cs2_bottom_xycorr = (peaks['area']
                             * (1 - peaks['area_fraction_top'])
                             / self.s2_xy_map(s2_positions, map_name=s2_bottom_map_name))

        # For electron lifetime corrections to the S2s,
        # use drift time computed using the main S1.

        elife_correction = np.exp(peaks['drift_time'] / self.elife)
        result['cs2_wo_timecorr'] = ((cs2_top_xycorr + cs2_bottom_xycorr) * elife_correction)

        for partition, func in self.regions.items():
            # partitioned SE and EE
            partition_mask = func(peaks['x'], peaks['y'])

            # Correct for SEgain and extraction efficiency
            seg_ee_corr = seg[partition] / avg_seg[partition] * ee[partition]

            # note that these are already masked!
            cs2_top_wo_elifecorr = cs2_top_xycorr[partition_mask] / seg_ee_corr
            cs2_bottom_wo_elifecorr = cs2_bottom_xycorr[partition_mask] / seg_ee_corr

            result['cs2_wo_elifecorr'][partition_mask] = cs2_top_wo_elifecorr + cs2_bottom_wo_elifecorr

            # cs2aft doesn't need elife/time corrections as they cancel
            result['cs2_area_fraction_top'][partition_mask] = cs2_top_wo_elifecorr / (
                        cs2_top_wo_elifecorr + cs2_bottom_wo_elifecorr)

            result['cs2'][partition_mask] = result['cs2_wo_elifecorr'][
                partition_mask] * elife_correction[partition_mask]
            result['cs2_bottom'][partition_mask] = cs2_bottom_wo_elifecorr * elife_correction[partition_mask]

        not_s2_mask = peaks['type'] != 2
        result['cs2_wo_timecorr'][not_s2_mask] = np.nan
        result['cs2_wo_elifecorr'][not_s2_mask] = np.nan
        result['cs2_area_fraction_top'][not_s2_mask] = np.nan
        result['cs2'][not_s2_mask] = np.nan
        result['z_obs_ms'][not_s2_mask] = np.nan
        result['cs2_bottom'][not_s2_mask] = np.nan
        result['s1_xyz_correction_factor'][not_s2_mask] = np.nan
        result['s1_rel_light_yield_correction_factor'][not_s2_mask] = np.nan
        return result



import warnings

import numpy as np
import strax
import straxen
import typing as ty
from functools import wraps
from straxen.corrections_services import corrections_w_file
from straxen.corrections_services import single_value_corrections
from straxen.corrections_services import arrays_corrections
from straxen.corrections_services import dict_corrections


export, __all__ = strax.exporter()



def correction_options(get_correction_function):
    """
    A wrapper function for functions here in the get_corrections module
    Search for special options like ["cmt_run_id", "prefix", "suffix"] and apply arg shuffling accordingly
    Example confs:
        ('cmt_run_id', cmt_run_id, 'to_pe_model', 'ONLINE', True)
        ('suffix', suffix, 'cmt_run_id', cmt_run_id, 'to_pe_model', 'ONLINE', True)

    :param get_correction_function: A function here in the get_corrections module
    :returns: The function wrapped with the option search
    """
    @wraps(get_correction_function)
    def correction_options_wrapped(run_id, conf, *arg):
        if isinstance(conf, tuple):
            set_prefix = ['prefix', False, None]
            set_suffix = ['suffix', False, None]
            set_cmt_run_id = ['cmt_run_id', False, None]

            for tag in [set_cmt_run_id, set_prefix, set_suffix]:
                if tag[0] in conf:
                    i_tag = conf.index(tag[0])
                    tag[:] = [tag[0], True, conf[i_tag + 1]]
                    conf = [item for i, item in enumerate(conf) if i not in [i_tag, i_tag + 1]]

            if len(conf) == 1:
                conf = conf[0]
            else:
                if set_prefix[1]:
                    conf[0] = set_prefix[2] + '_' + conf[0]
                if set_suffix[1]:
                    conf[0] = conf[0] + '_' + set_suffix[2]
                if set_cmt_run_id[1]:
                    run_id = set_cmt_run_id[2]

                conf = tuple(conf)

        # Else use the get corrections as they are
        return get_correction_function(run_id, conf, *arg)

    return correction_options_wrapped


@export
@correction_options
def get_correction_from_cmt(run_id, conf):
    """
    Get correction from CMT general format is
    conf = ('correction_name', 'version', True)
    where True means looking at nT runs, e.g. 
    get_correction_from_cmt(run_id, conf[:2])
    special cases:
    version can be replaced by constant int, float or array
    when user specify value(s)
    :param run_id: run id from runDB
    :param conf: configuration 
    :return: correction value(s)
    """
 
    if isinstance(conf, str):
        # Legacy support for pax files
        return conf

    elif isinstance(conf, tuple) and len(conf) == 2:
        model_conf, cte_value = conf[:2]

        # special case constant to_pe values should be covered by legacy protocols?
        from straxen.legacy.xenon1t_url_configs import FIXED_TO_PE
        if model_conf in FIXED_TO_PE:
            warnings.warn(f'Don\'t load like this, but via legacy config')
            correction = FIXED_TO_PE[model_conf]
            return correction

        # special case constant single value or list of values.
        elif 'constant' in model_conf:
            if not isinstance(cte_value, (float, int, str, list, tuple)):
                raise ValueError(f"User specify a model type {model_conf} "
                                 "and should provide a number or list of numbers. Got: "
                                 f"{type(cte_value)}")
            correction = cte_value
            return correction

    elif isinstance(conf, tuple) and len(conf) == 3:
        model_conf, global_version, is_nt = conf[:3]
        cmt = straxen.CorrectionsManagementServices(is_nt=is_nt)
        correction = cmt.get_corrections_config(run_id, conf[:2])
        if correction.size == 0:
            raise ValueError(f"Could not find a value for {model_conf} "
                             f"please check it is implemented in CMT. ")

        if model_conf in corrections_w_file: # file's name (maps, NN, etc) 
            correction = ' '.join(map(str, correction))
            return correction

        if model_conf in dict_corrections:
            return correction[0]

        elif model_conf in single_value_corrections:
            if 'samples' in model_conf: # int baseline samples, etc
                return int(correction)
            else:
                return float(correction) # float elife, drift velocity, etc
        
        elif model_conf in arrays_corrections:
            np_correction = correction.reshape(correction.size)
            np_correction = np_correction.astype(np.int16)  # not sure if straxen can handle dtype:object therefore specify dtype
            return np_correction
        
        return correction
    
    else:
        raise ValueError("Wrong configuration. "
                         "Please use the following format: "
                         "(config->str, model_config->str or number, is_nT->bool) "
                         f"User specify {conf} please modify")


@export
def get_cmt_resource(run_id, conf, fmt=''):
    """
    Get resource with CMT correction file name
    """
    return straxen.get_resource(get_correction_from_cmt(run_id, conf), fmt=fmt)


@export
def is_cmt_option(config):
    """
    Check if the input configuration is cmt style.
    """
    return _is_cmt_option(None, config)


@correction_options
def _is_cmt_option(run_id, config):
    # Compatibilty with URLConfig
    if isinstance(config, str) and "cmt://" in config:
        return True
    is_cmt = (isinstance(config, tuple)
              and len(config)==3
              and isinstance(config[0], str)
              and isinstance(config[1], (str, int, float))
              and isinstance(config[2], bool))
    
    return is_cmt


def get_cmt_options(context: strax.Context) -> ty.Dict[str, ty.Dict[str, tuple]]:
    """
    Function which loops over all plugin configs and returns dictionary
    with option name as key and a nested dict of CMT correction name and strax option as values.

    :param context: Context with registered plugins.
    """

    cmt_options = {}
    runid_test_str = '0000'

    for data_type, plugin in context._plugin_class_registry.items():
        for option_key, option in plugin.takes_config.items():
            if option_key in cmt_options:
                # let's not do work twice if needed by > 1 plugin
                continue

            if (option_key in context.config and
                    is_cmt_option(context.config[option_key])):
                opt = context.config[option_key]
            elif is_cmt_option(option.default):
                opt = option.default
            else:
                continue

            # check if it's a URLConfig
            if isinstance(opt, str) and 'cmt://' in opt:
                before_cmt, cmt, after_cmt = opt.partition('cmt://')
                p = context._get_plugins((data_type,), runid_test_str)[data_type]
                context._set_plugin_config(p, runid_test_str, tolerant=False)
                del p.run_id

                p.config[option_key] = after_cmt
                try:
                    correction_name = getattr(p, option_key)
                except AttributeError:
                    # make sure the correction name does not depend on runid
                    raise RuntimeError("Correction names should not depend on runids! "
                                       f"Please check your option for {option_key}")

                # if there is no other protocol being called before cmt,
                # we will get a string back including the query part
                if option.QUERY_SEP in correction_name:
                    correction_name, _ = option.split_url_kwargs(correction_name)
                cmt_options[option_key] = {'correction': correction_name,
                                           'strax_option': opt,
                                           }

            else:
                cmt_options[option_key] = {'correction': opt[0],
                                           'strax_option': opt,
                                           }
    return cmt_options



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class EventShadow(strax.Plugin):
    """
    This plugin can calculate shadow for main S1 and main S2 in events.
    It also gives the position information of the previous peaks.
    References:
        * v0.1.4 reference: xenon:xenonnt:ac:prediction:shadow_ambience
    """
    __version__ = '0.1.4'
    depends_on = ('event_basics', 'peak_basics', 'peak_shadow')
    provides = 'event_shadow'

    def infer_dtype(self):
        dtype = []
        for main_peak, main_peak_desc in zip(['s1_', 's2_'], ['main S1', 'main S2']):
            # previous S1 can only cast time shadow, previous S2 can cast both time & position shadow
            for key in ['s1_time_shadow', 's2_time_shadow', 's2_position_shadow']:
                type_str, tp_desc, _ = key.split('_')
                dtype.append(((f'largest {tp_desc} shadow casting from previous {type_str} to {main_peak_desc} [PE/ns]',
                               f'{main_peak}shadow_{key}'), np.float32))
                dtype.append(((
                                  f'time difference from the previous {type_str} casting largest {tp_desc} shadow to {main_peak_desc} [ns]',
                                  f'{main_peak}dt_{key}'), np.int64))
                # Only previous S2 peaks have (x,y)
                if 's2' in key:
                    dtype.append(((f'x of previous s2 peak casting largest {tp_desc} shadow on {main_peak_desc} [cm]',
                                   f'{main_peak}x_{key}'), np.float32))
                    dtype.append(((f'y of previous s2 peak casting largest {tp_desc} shadow on {main_peak_desc} [cm]',
                                   f'{main_peak}y_{key}'), np.float32))
                # Only time shadow gives the nearest large peak
                if 'time' in key:
                    dtype.append(
                        ((f'time difference from the nearest previous large {type_str} to {main_peak_desc} [ns]',
                          f'{main_peak}nearest_dt_{type_str}'), np.int64))
            # Also record the PDF of HalfCauchy when calculating S2 position shadow
            dtype.append(((f'PDF describing correlation between previous s2 and {main_peak_desc}',
                           f'{main_peak}pdf_s2_position_shadow'), np.float32))
        dtype += strax.time_fields
        return dtype

    @staticmethod
    def set_nan_defaults(result):
        """
        When constructing the dtype, take extra care to set values to
        np.Nan / -1 (for ints) as 0 might have a meaning
        """
        for field in result.dtype.names:
            if np.issubdtype(result.dtype[field], np.integer):
                result[field][:] = -1
            else:
                result[field][:] = np.nan

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        result = np.zeros(len(events), self.dtype)

        self.set_nan_defaults(result)

        # 1. Assign peaks features to main S1 and main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            res_i = result[event_i]
            # Fetch the features of main S1 and main S2
            for idx, main_peak in zip([event['s1_index'], event['s2_index']], ['s1_', 's2_']):
                if idx >= 0:
                    for key in ['s1_time_shadow', 's2_time_shadow', 's2_position_shadow']:
                        type_str = key.split('_')[0]
                        res_i[f'{main_peak}shadow_{key}'] = sp[f'shadow_{key}'][idx]
                        res_i[f'{main_peak}dt_{key}'] = sp[f'dt_{key}'][idx]
                        if 'time' in key:
                            res_i[f'{main_peak}nearest_dt_{type_str}'] = sp[f'nearest_dt_{type_str}'][idx]
                        if 's2' in key:
                            res_i[f'{main_peak}x_{key}'] = sp[f'x_{key}'][idx]
                            res_i[f'{main_peak}y_{key}'] = sp[f'y_{key}'][idx]
                    # Record the PDF of HalfCauchy
                    res_i[f'{main_peak}pdf_s2_position_shadow'] = sp['pdf_s2_position_shadow'][idx]

        # 2. Set time and endtime for events
        result['time'] = events['time']
        result['endtime'] = strax.endtime(events)
        return result



import strax
import numpy as np


class EventwBayesClass(strax.Plugin):
    """
    Append at event level the posterior probability for an
    S1, S2, alt_S1 and alt_S2
    """
    provides = 'event_w_bayes_class'
    depends_on = ('peak_classification_bayes', 'event_basics')
    data_kind = 'events'
    __version__ = '0.0.1'

    def infer_dtype(self):

        dtype = []
        dtype += strax.time_fields
        for name in ['s1', 's2', 'alt_s1', 'alt_s2']:
            dtype += [(f'{name}_ln_prob_s1', np.float32, f'Given an {name}, s1 ln probability')]
            dtype += [(f'{name}_ln_prob_s2', np.float32, f'Given an {name}, s2 ln probability')]

        return dtype

    def compute(self, peaks, events):

        result = np.empty(len(events), dtype=self.dtype)

        for name in ['s1', 's2', 'alt_s1', 'alt_s2']:
            result[f'{name}_ln_prob_s1'] = np.nan
            result[f'{name}_ln_prob_s2'] = np.nan
            # Select peaks based on their start time
            mask = np.in1d(peaks['time'], events[f'{name}_time'])
            mask_ev = np.in1d(events[f'{name}_time'], peaks['time'])
            result[f'{name}_ln_prob_s1'][mask_ev] = peaks['ln_prob_s1'][mask]
            result[f'{name}_ln_prob_s2'][mask_ev] = peaks['ln_prob_s2'][mask]
            result['time'] = events['time']
            result['endtime'] = events['endtime']

        return result



import strax
import straxen

from straxen.plugins.events_nv.events_nv import veto_event_dtype, nVETOEvents

export, __all__ = strax.exporter()


@export
class muVETOEvents(nVETOEvents):
    """Plugin which computes the boundaries of veto events.
    """
    depends_on = 'hitlets_mv'
    provides = 'events_mv'
    data_kind = 'events_mv'

    compressor = 'zstd'
    child_plugin = True

    __version__ = '0.0.1'
    events_seen = 0

    event_left_extension_mv = straxen.URLConfig(
        default=0,
        track=True,
        type=int,
        child_option=True,
        parent_option_name='event_left_extension_nv',
        help='Extends event window this many [ns] to the left.'
    )
    event_resolving_time_mv = straxen.URLConfig(
        default=300,
        track=True,
        type=int,
        child_option=True,
        parent_option_name='event_resolving_time_nv',
        help='Resolving time for window coincidence [ns].'
    )
    event_min_hits_mv = straxen.URLConfig(
        default=3,
        track=True,
        type=int,
        child_option=True,
        parent_option_name='event_min_hits_nv',
        help='Minimum number of fully confined hitlets to define an event.'
    )

    def infer_dtype(self):
        self.name_event_number = 'event_number_mv'
        self.channel_range = self.channel_map['mv']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_event_dtype(self.name_event_number, self.n_channel)

    def get_window_size(self):
        return self.event_left_extension_mv + self.event_resolving_time_mv + 1

    def compute(self, hitlets_mv, start, end):
        return super().compute(hitlets_mv, start, end)



from immutabledict import immutabledict
import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class EventAreaPerChannel(strax.Plugin):
    """
    Simple plugin that provides area per channel for main and alternative S1/S2 in the event.
    """
    
    depends_on = ('event_basics', 'peaks')
    provides = ("event_area_per_channel", "event_n_channel")
    data_kind = immutabledict(zip(provides, ("events", "events")))
    __version__ = '0.1.1'

    compressor = 'zstd'
    save_when = immutabledict({
        "event_area_per_channel": strax.SaveWhen.EXPLICIT,
        "event_n_channel": strax.SaveWhen.ALWAYS,
    })

    n_top_pmts = straxen.URLConfig(default=straxen.n_top_pmts, type=int, help="Number of top PMTs")

    def infer_dtype(self):
        # setting data type from peak dtype
        pfields_ = self.deps['peaks'].dtype_for('peaks').fields
        # populating data type
        infoline = {'s1': 'main S1',
                    's2': 'main S2',
                    'alt_s1': 'alternative S1',
                    'alt_s2': 'alternative S2',
                   }
        dtype = []
        # populating APC
        ptypes = ['s1', 's2', 'alt_s1', 'alt_s2']
        for type_ in ptypes:
            dtype += [((f'Area per channel for {infoline[type_]}', f'{type_}_area_per_channel'),
                     pfields_['area_per_channel'][0])]
            dtype += [((f'Length of the interval in samples for {infoline[type_]}', f'{type_}_length'),
                     pfields_['length'][0])]
            dtype += [((f'Width of one sample for {infoline[type_]} [ns]', f'{type_}_dt'),
                     pfields_['dt'][0])]
        # populating S1 n channel properties
        n_channel_dtype = [(("Main S1 count of contributing PMTs", "s1_n_channels"),
                  np.int16),
                 (("Main S1 top count of contributing PMTs", "s1_top_n_channels"),
                  np.int16),
                 (("Main S1 bottom count of contributing PMTs", "s1_bottom_n_channels"),
                  np.int16),
                 ]
        return {
            "event_area_per_channel": dtype + n_channel_dtype + strax.time_fields,
            "event_n_channel": n_channel_dtype + strax.time_fields,
        }

    def compute(self, events, peaks):
        area_per_channel = np.zeros(len(events), self.dtype["event_area_per_channel"])
        area_per_channel['time'] = events['time']
        area_per_channel['endtime'] = strax.endtime(events)
        n_channel = np.zeros(len(events), self.dtype["event_n_channel"])
        n_channel['time'] = events['time']
        n_channel['endtime'] = strax.endtime(events)

        split_peaks = strax.split_by_containment(peaks, events)
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for type_ in ['s1', 's2', 'alt_s1', 'alt_s2']:
                type_index = event[f'{type_}_index']
                if type_index != -1:
                    type_area_per_channel = sp['area_per_channel'][type_index]
                    area_per_channel[f'{type_}_area_per_channel'][event_i] = type_area_per_channel
                    area_per_channel[f'{type_}_length'][event_i] = sp['length'][type_index]
                    area_per_channel[f'{type_}_dt'][event_i] = sp['dt'][type_index]
                    if type_ == 's1':
                        area_per_channel['s1_n_channels'][event_i] = (
                            type_area_per_channel > 0).sum()
                        area_per_channel['s1_top_n_channels'][event_i] = (
                            type_area_per_channel[:self.config['n_top_pmts']] > 0).sum()
                        area_per_channel['s1_bottom_n_channels'][event_i] = (
                            type_area_per_channel[self.config['n_top_pmts']:] > 0).sum()
        for field in ['s1_n_channels', 's1_top_n_channels', 's1_bottom_n_channels']:
            n_channel[field] = area_per_channel[field]
        result = {
            "event_area_per_channel": area_per_channel,
            "event_n_channel": n_channel,
        }
        return result



import strax
import straxen
from straxen.plugins.defaults import MV_PREAMBLE, NV_HIT_DEFAULTS, MV_HIT_DEFAULTS


export, __all__ = strax.exporter()


@export
class nVETOPulseProcessing(strax.Plugin):
    """
    nVETO equivalent of pulse processing. The following steps are
    applied:

        1. Flip, baseline and integrate waveforms.
        2. Find hits and apply ZLE
        3. Remove empty fragments.
    """
    __version__ = '0.0.8'

    parallel = 'process'
    rechunk_on_save = False
    compressor = 'zstd'
    save_when = strax.SaveWhen.TARGET

    depends_on = 'raw_records_coin_nv'
    provides = 'records_nv'
    data_kind = 'records_nv'

    save_outside_hits_nv = straxen.URLConfig(
        default=NV_HIT_DEFAULTS['save_outside_hits_nv'], track=True, infer_type=False,
        help='Save (left, right) samples besides hits; cut the rest')

    hit_min_amplitude_nv = straxen.URLConfig(
        infer_type=False,
        default=NV_HIT_DEFAULTS['hit_min_amplitude_nv'], track=True,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_nveto_pmts, or a number, '
             'or a string like "pmt_commissioning_initial" which means calling '
             'hitfinder_thresholds.py, '
             'or a tuple like (correction=str, version=str, nT=boolean), '
             'which means we are using cmt.')

    baseline_samples_nv = straxen.URLConfig(
        infer_type=False,
        default='cmt://baseline_samples_nv?version=ONLINE&run_id=plugin.run_id', track=True,
        help='Number of samples to use at the start of the pulse to determine '
             'the baseline')

    def setup(self):
        self.baseline_samples = self.baseline_samples_nv
        self.hit_thresholds = self.hit_min_amplitude_nv

    def infer_dtype(self):
        record_length = strax.record_length_from_dtype(
            self.deps['raw_records_coin_nv'].dtype_for('raw_records_coin_nv'))
        dtype = strax.record_dtype(record_length)
        return dtype

    def compute(self, raw_records_coin_nv):
        # Do not trust in DAQ + strax.baseline to leave the
        # out-of-bounds samples to zero.
        r = strax.raw_to_records(raw_records_coin_nv)
        del raw_records_coin_nv

        r = strax.sort_by_time(r)
        strax.zero_out_of_bounds(r)
        strax.baseline(r,
                       baseline_samples=self.baseline_samples,
                       flip=True)

        strax.integrate(r)

        strax.zero_out_of_bounds(r)

        hits = strax.find_hits(r, min_amplitude=self.hit_thresholds)

        le, re = self.save_outside_hits_nv
        r = strax.cut_outside_hits(r, hits, left_extension=le, right_extension=re)
        strax.zero_out_of_bounds(r)

        return r



import numpy as np
import strax
import straxen
from straxen.plugins.peaklets.peaklets import Peaklets
from straxen.plugins.defaults import HE_PREAMBLE

export, __all__ = strax.exporter()


@export
class PeakletsHighEnergy(Peaklets):
    __doc__ = HE_PREAMBLE + Peaklets.__doc__
    depends_on = 'records_he'
    provides = 'peaklets_he'
    data_kind = 'peaklets_he'
    __version__ = '0.1.0'
    child_plugin = True
    save_when = strax.SaveWhen.TARGET

    n_he_pmts = straxen.URLConfig(
        track=False, default=752, infer_type=False,
        help="Maximum channel of the he channels")

    he_channel_offset = straxen.URLConfig(
        track=False, default=500, infer_type=False,
        help="Minimum channel number of the he channels")

    le_to_he_amplification = straxen.URLConfig(
        default=20, track=True, infer_type=False,
        help="Difference in amplification between low energy and high energy channels")

    peak_min_pmts_he = straxen.URLConfig(
        default=2, infer_type=False,
        child_option=True, parent_option_name='peak_min_pmts',
        track=True,
        help="Minimum number of contributing PMTs needed to define a peak")

    saturation_correction_on_he = straxen.URLConfig(
        default=False, infer_type=False,
        child_option=True, parent_option_name='saturation_correction_on',
        track=True,
        help='On off switch for saturation correction for High Energy'
             ' channels')

    hit_min_amplitude_he = straxen.URLConfig(
        default='cmt://hit_thresholds_he?version=ONLINE&run_id=plugin.run_id', track=True, infer_type=False,
        help='Minimum hit amplitude in ADC counts above baseline. '
             'Specify as a tuple of length n_tpc_pmts, or a number,'
             'or a string like "pmt_commissioning_initial" which means calling'
             'hitfinder_thresholds.py'
             'or a tuple like (correction=str, version=str, nT=boolean),'
             'which means we are using cmt.'
    )

    # We cannot, we only have the top array, so should not.
    sum_waveform_top_array = False

    @property
    def n_tpc_pmts(self):
        # Have to hack the url config to avoid nasty numba errors for the main Peaklets plugin
        return self.n_he_pmts

    def infer_dtype(self):
        return strax.peak_dtype(n_channels=self.n_he_pmts, digitize_top=self.sum_waveform_top_array)

    def setup(self):
        self.to_pe = self.gain_model
        buffer_pmts = np.zeros(self.he_channel_offset)
        self.to_pe = np.concatenate((buffer_pmts, self.to_pe))
        self.to_pe *= self.le_to_he_amplification
        self.hit_thresholds = self.hit_min_amplitude_he
        self.channel_range = self.channel_map['he']

    def compute(self, records_he, start, end):
        result = super().compute(records_he, start, end)
        return result['peaklets']



import strax
import straxen
from straxen.holoviews_utils import nVETOEventDisplay
from straxen.holoviews.holoviews_peak_data import PlotPeaksTPC
from straxen.holoviews.holoviews_pmt_array import PlotPMTArrayTPC
from straxen.holoviews.holoviews_inspector import EventStats
from straxen.holoviews.holoviews_tpc_event_display import InteractiveTPCEventDisplay
import holoviews as hv
import panel as pn
import numpy as np
from tempfile import TemporaryDirectory
import os
import pandas as pd


_nveto_pmt_dummy = {'channel': list(range(2000, 2120)),
                    'x': list(range(120)),
                    'y': list(range(120)),
                    'z': list(range(120)),
                    }
_dummy_map = pd.DataFrame(_nveto_pmt_dummy).to_records()


def test_hitlets_to_hv_points():
    hit = np.zeros(1, dtype=strax.hit_dtype)
    hit['time'] = 10
    hit['length'] = 2
    hit['dt'] = 1
    hit['channel'] = 2000
    hit['area'] = 1

    nvd = nVETOEventDisplay(pmt_map=_dummy_map)
    points = nvd.hitlets_to_hv_points(hit, t_ref=0)

    m = [hit[key] == points.data[key] for key in hit.dtype.names if
         key in points.data.columns.values]
    assert np.all(m), 'Data has not been converted corretly into hv.Points.'


def test_hitlet_matrix():
    hit = np.zeros(1, dtype=strax.hit_dtype)
    hit['time'] = 10
    hit['length'] = 2
    hit['dt'] = 1
    hit['channel'] = 2000
    hit['area'] = 1

    nvd = nVETOEventDisplay(pmt_map=_dummy_map)
    hit_m = nvd.plot_hitlet_matrix(hitlets=hit)

    with TemporaryDirectory() as d:
        # Have to store plot to make sure it is rendered
        hv.save(hit_m, os.path.join(d, 'hitlet_matrix.html'))


def test_plot_nveto_pattern():
    hit = np.zeros(1, dtype=strax.hit_dtype)
    hit['channel'] = 2000
    hit['area'] = 1

    nvd = nVETOEventDisplay(pmt_map=_dummy_map)
    pmt_plot = nvd.plot_nveto(hitlets=hit)
    with TemporaryDirectory() as d:
        # Have to store plot to make sure it is rendered
        hv.save(pmt_plot, os.path.join(d, 'hitlet_matrix.html'))


def test_nveto_event_display():
    hit = np.zeros(1, dtype=strax.hit_dtype)
    hit['time'] = 10
    hit['length'] = 2
    hit['dt'] = 1
    hit['channel'] = 2000
    hit['area'] = 1

    dtype = straxen.plugins.events_nv.veto_event_dtype()
    dtype += straxen.plugins.event_positions_nv.veto_event_positions_dtype()[2:]
    event = np.zeros(1, dtype=dtype)
    event['time'] = hit['time']
    event['endtime'] = hit['time'] + 40
    event['area'] = hit['area']

    nvd = nVETOEventDisplay(event, hit, pmt_map=_dummy_map, run_id='014986')
    dispaly = nvd.plot_event_display()

    with TemporaryDirectory() as d:
        # Have to store plot to make sure it is rendered
        pn.io.save.save(dispaly, os.path.join(d, 'hitlet_matrix.html'))


def test_array_to_df_and_make_sliders():
    dtype = (straxen.plugins.events_nv.veto_event_dtype()
             + straxen.plugins.event_positions_nv.veto_event_positions_dtype()[2:])
    evt = np.zeros(1, dtype)

    nvd = nVETOEventDisplay(pmt_map=_dummy_map)
    df = straxen.convert_array_to_df(evt)

    nvd._make_sliders_and_tables(df)


def test_static_detector_plots():
    tpc = straxen.holoviews_utils.plot_tpc_circle(straxen.cryostat_outer_radius)
    diffuer_balls = straxen.holoviews_utils.plot_diffuser_balls_nv()
    nveto_reflector = straxen.holoviews_utils.plot_nveto_reflector()

    with TemporaryDirectory() as d:
        # Have to store plot to make sure it is rendered
        hv.save(tpc * diffuer_balls * nveto_reflector, os.path.join(d, 'hitlet_matrix.html'))


def test_pmt_array_plot():
    """Test if PMT array can be plotted for the interactive
    display.
    """
    dummy_gains = np.ones(494)
    dummy_gains[100:120] = 0

    dummy_peak = np.ones(1, dtype=strax.peak_dtype(n_channels=straxen.n_tpc_pmts))

    test = PlotPMTArrayTPC(gains=dummy_gains, top_pmt_array=True)
    test.plot_pmt_array(dummy_peak[0])

    test = PlotPMTArrayTPC(top_pmt_array=False)
    test.plot_pmt_array(dummy_peak[0])


def test_tpc_display_components():
    dummy_peak = np.zeros(3, dtype=strax.peak_dtype(494))
    dummy_peak['time'] = [0, 20, 40]
    dummy_peak['dt'] = 1
    dummy_peak['data'][0,:10] = np.arange(10)
    dummy_peak['length'] = 10
    dummy_peak['area_per_channel'] = 1

    peak_plotter = PlotPeaksTPC()
    peak_plotter.plot_peaks(dummy_peak)

    array_plotter = PlotPMTArrayTPC()
    array_plotter.plot_pmt_array(dummy_peak[0])

    st = straxen.contexts.xenonnt_online(_database_init=False)
    p = st.get_single_plugin('0', 'event_info')
    dummy_events = np.zeros(10, p.dtype)
    dummy_events[0]['endtime'] = 10

    dummy_peak['type'] = [0, 1, 2]
    display = InteractiveTPCEventDisplay(dummy_events, dummy_peak, plot_alt_peaks=True)
    event_display = display.event_display(inspector_config={'x_bins': (50, 0, 200),
                                            'y_bins': (50, 0, 200),
                                            'y_field': 's1_rise_time'})

    with TemporaryDirectory() as d:
        event_display.save(os.path.join(d, 'display_test.html'))

def test_inspector_plot():
    st = straxen.contexts.xenonnt_online(_database_init=False)
    p = st.get_single_plugin('0', 'event_info')
    dummy_events = np.zeros(10, p.dtype)
    dummy_events[0]['endtime'] = 10
    dummy_events['cs1'][::2] = 1
    dummy_events['cs1'][::2] = 2
    dummy_events['cs1'][1::2] = 2
    dummy_events['cs1'][1::2] = 4

    inspector = EventStats(dummy_events,
                           )
    plot, widget =inspector.interactive_event_stats()
    with TemporaryDirectory() as d:
        plot.save(os.path.join(d, 'inspector_plottest.html'))

    inspector = EventStats(dummy_events,
                           exclude_fields=['s1_area',],
                           default_config={'x_bins': (10, 0, 10),
                                           'y_bins': (10, 0, 10),
                                           'y_field': 's1_rise_time'}
                           )
    plot, widget = inspector.interactive_event_stats()

    # Trigger replotting:
    w = widget[0]
    w.value = 1

    with TemporaryDirectory() as d:
        plot.save(os.path.join(d, 'inspector_plottest.html'))



"""For all of the context, do a quick check to see that we are able to search
a field (i.e. can build the dependencies in the context correctly)
See issue #233 and PR #236"""
from straxen.contexts import xenon1t_dali, xenon1t_led, fake_daq, demo
from straxen.contexts import xenonnt_led, xenonnt_online, xenonnt
import straxen
import tempfile
import os
import unittest


##
# XENONnT
##


def test_xenonnt_online():
    st = xenonnt_online(_database_init=False)
    st.search_field('time')


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_xenonnt_online_with_online_frontend():
    st = xenonnt_online(include_online_monitor=True)
    for sf in st.storage:
        if 'OnlineMonitor' == sf.__class__.__name__:
            break
    else:
        raise ValueError(f"Online monitor not in {st.storage}")


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_xenonnt_online_rucio_local():
    st = xenonnt_online(include_rucio_local=True, _rucio_local_path='./test')
    for sf in st.storage:
        if 'RucioLocalFrontend' == sf.__class__.__name__:
            break
    else:
        raise ValueError(f"Online monitor not in {st.storage}")


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_xennonnt():
    st = xenonnt(_database_init=False)
    st.search_field('time')


def test_xenonnt_led():
    st = xenonnt_led(_database_init=False)
    st.search_field('time')


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_nt_is_nt_online():
    # Test that nT and nT online are the same
    st_online = xenonnt_online(_database_init=False)

    st = xenonnt(_database_init=False)
    for plugin in st._plugin_class_registry.keys():
        print(f'Checking {plugin}')
        nt_key = st.key_for('0', plugin)
        nt_online_key = st_online.key_for('0', plugin)
        assert str(nt_key) == str(nt_online_key)


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_offline():
    """
    Let's try and see which CMT versions are compatible with this straxen
    version
    """
    cmt = straxen.CorrectionsManagementServices()
    cmt_versions = list(cmt.global_versions)[::-1]
    print(cmt_versions)
    success_for = []
    for global_version in cmt_versions:
        try:
            xenonnt(global_version)
            success_for.append(global_version)
        except straxen.CMTVersionError:
            pass
    print(f'This straxen version works with {success_for} but is '
          f'incompatible with {set(cmt_versions)-set(success_for)}')

    test = unittest.TestCase()
    # We should always work for one offline and the online version
    test.assertTrue(len(success_for) >= 2)


##
# XENON1T
##


def test_xenon1t_dali():
    st = xenon1t_dali()
    st.search_field('time')


def test_demo():
    """
    Test the demo context. Since we download the folder to the current
        working directory, make sure we are in a tempfolder where we
        can write the data to
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            print("Temporary directory is ", temp_dir)
            os.chdir(temp_dir)
            st = demo()
            st.search_field('time')
        # On windows, you cannot delete the current process'
        # working directory, so we have to chdir out first.
        finally:
            os.chdir('..')


def test_fake_daq():
    st = fake_daq()
    st.search_field('time')


def test_xenon1t_led():
    st = xenon1t_led()
    st.search_field('time')

##
# WFSim
##


# Simulation contexts are only tested when special flags are set

@unittest.skipIf('ALLOW_WFSIM_TEST' not in os.environ,
                 "if you want test wfsim context do `export 'ALLOW_WFSIM_TEST'=1`")
class TestSimContextNT(unittest.TestCase):
    @staticmethod
    def context(*args, **kwargs):
        kwargs.setdefault('cmt_version', 'global_ONLINE')
        return straxen.contexts.xenonnt_simulation(*args, **kwargs)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_nt_sim_context_main(self):
        st = self.context(cmt_run_id_sim='008000')
        st.search_field('time')

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_nt_sim_context_alt(self):
        """Some examples of how to run with a custom WFSim context"""
        self.context(cmt_run_id_sim='008000', cmt_run_id_proc='008001')
        self.context(cmt_run_id_sim='008000',
                     cmt_option_overwrite_sim={'elife': 1e6})

        self.context(cmt_run_id_sim='008000',
                     overwrite_fax_file_sim={'elife': 1e6})

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_nt_diverging_context_options(self):
        """
        Test diverging options. Idea is that you can use different
        settings for processing and generating data, should have been
        handled by RawRecordsFromWFsim but is now hacked into the
        xenonnt_simulation context

        Just to show how convoluted this syntax for the
        xenonnt_simulation context / CMT is...
        """
        self.context(cmt_run_id_sim='008000',
                     cmt_option_overwrite_sim={'elife': ('elife_constant', 1e6, True)},
                     cmt_option_overwrite_proc={'elife': ('elife_constant', 1e5, True)},
                     overwrite_from_fax_file_proc=True,
                     overwrite_from_fax_file_sim=True,
                     _config_overlap={'electron_lifetime_liquid': 'elife'},
                     )

    def test_nt_sim_context_bad_inits(self):
        with self.assertRaises(RuntimeError):
            self.context(cmt_run_id_sim=None, cmt_run_id_proc=None,)

@unittest.skipIf('ALLOW_WFSIM_TEST' not in os.environ,
                 "if you want test wfsim context do `export 'ALLOW_WFSIM_TEST'=1`")
def test_sim_context():
    st = straxen.contexts.xenon1t_simulation()


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
@unittest.skipIf(
    "ALLOW_WFSIM_TEST" not in os.environ,
    "if you want test wfsim context do `export 'ALLOW_WFSIM_TEST'=1`",
)
def test_sim_offline_context():
    st = straxen.contexts.xenonnt_simulation_offline(
        run_id="026000",
        global_version="global_v11",
        fax_config="fax_config_nt_sr0_v4.json",
    )


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_offline():
    st = xenonnt('latest')
    st.provided_dtypes()



import strax
import straxen
from straxen.plugins.peaks._peak_positions_base import PeakPositionsBaseNT


export, __all__ = strax.exporter()


@export
class PeakPositionsGCN(PeakPositionsBaseNT):
    """Graph Convolutional Network (GCN) neural net for position reconstruction"""
    provides = "peak_positions_gcn"
    algorithm = "gcn"
    __version__ = '0.0.1'

    tf_model_gcn = straxen.URLConfig(
        default=f'tf://'
                f'resource://'
                f'cmt://{algorithm}_model'
                f'?version=ONLINE'
                f'&run_id=plugin.run_id'
                f'&fmt=abs_path',
        help='GCN model. Should be opened using the "tf" descriptor. '
             'Set to "None" to skip computation',
        cache=3,
    )



import numpy as np
import numba
from scipy.stats import halfcauchy
from .peak_ambience import distance_in_xy, _quick_assign
import strax
import straxen

export, __all__ = strax.exporter()


@export
class PeakShadow(strax.OverlapWindowPlugin):
    """
    This plugin can find and calculate the time & position shadow
    from previous peaks in time.
    It also gives the area and (x,y) of the previous peaks.
    References:
        * v0.1.5 reference: xenon:xenonnt:ac:prediction:shadow_ambience
    """

    __version__ = '0.1.6'

    depends_on = ('peak_basics', 'peak_positions')
    provides = 'peak_shadow'
    save_when = strax.SaveWhen.EXPLICIT

    shadow_time_window_backward = straxen.URLConfig(
        default=int(1e9),
        type=int,
        track=True,
        help='Search for peaks casting time & position shadow in this time window [ns]'
    )

    shadow_threshold = straxen.URLConfig(
        default={'s1_time_shadow': 1e3,
                 's2_time_shadow': 1e4,
                 's2_position_shadow': 1e4},
        type=dict,
        track=True,
        help='Only take S1/S2s larger than this into account when calculating Shadow [PE]'
    )

    shadow_deltatime_exponent = straxen.URLConfig(
        default=-1.0,
        type=float,
        track=True,
        help='The exponent of delta t when calculating shadow'
    )

    shadow_sigma_and_baseline = straxen.URLConfig(
        default=[15.220, 0.036],
        type=list,
        track=True,
        help='Fitted position correlation sigma[cm*PE^0.5] and baseline[cm] using in position shadow'
    )

    def get_window_size(self):
        return 10 * self.shadow_time_window_backward

    def infer_dtype(self):
        s1_time_shadow_dtype = []
        s2_time_shadow_dtype = []
        s2_position_shadow_dtype = []
        nearest_dtype = []
        # We have time shadow(S2/dt) and position shadow(S2/dt*p(s))
        # previous S1 can only cast time shadow, previous S2 can cast both time & position shadow
        for key, dtype in zip(['s1_time_shadow', 's2_time_shadow', 's2_position_shadow'],
                              [s1_time_shadow_dtype, s2_time_shadow_dtype, s2_position_shadow_dtype]):
            type_str, tp_desc, _ = key.split('_')
            dtype.append(((f'previous large {type_str} casted largest {tp_desc} shadow [PE/ns]',
                           f'shadow_{key}'), np.float32))
            dtype.append \
                (((f'time difference to the previous large {type_str} peak casting largest {tp_desc} shadow [ns]',
                   f'dt_{key}'), np.int64))
            # Only previous S2 peaks have (x,y)
            if 's2' in key:
                dtype.append(((f'x of previous large s2 peak casting largest {tp_desc} shadow [cm]',
                               f'x_{key}'), np.float32))
                dtype.append(((f'y of previous large s2 peak casting largest {tp_desc} shadow [cm]',
                               f'y_{key}'), np.float32))
            # Only time shadow gives the nearest large peak
            if 'time' in key:
                dtype.append(((f'time difference to the nearest previous large {type_str} [ns]',
                               f'nearest_dt_{type_str}'), np.int64))
        # Also record the PDF of HalfCauchy when calculating S2 position shadow
        s2_position_shadow_dtype.append((('PDF describing correlation to the previous large s2',
                                          'pdf_s2_position_shadow'), np.float32))

        dtype = s1_time_shadow_dtype + s2_time_shadow_dtype + s2_position_shadow_dtype + nearest_dtype + strax.time_fields
        return dtype

    @property
    def shadowdtype(self):
        dtype = []
        dtype += [('shadow', np.float32), ('dt', np.int64)]
        dtype += [('x', np.float32), ('y', np.float32)]
        dtype += [('nearest_dt', np.int64)]
        return dtype

    def compute(self, peaks):
        argsort = np.argsort(peaks['center_time'], kind='mergesort')
        _peaks = np.sort(peaks, order='center_time')
        result = np.zeros(len(peaks), self.dtype)
        _quick_assign(argsort, result, self.compute_shadow(peaks, _peaks))
        return result

    def compute_shadow(self, peaks, current_peak):
        # 1. Define time window for each peak, we will find previous peaks within these time windows
        roi_shadow = np.zeros(len(current_peak), dtype=strax.time_fields)
        roi_shadow['time'] = current_peak['center_time'] - self.shadow_time_window_backward
        roi_shadow['endtime'] = current_peak['center_time']

        # 2. Calculate S2 position shadow, S2 time shadow, and S1 time shadow
        result = np.zeros(len(current_peak), self.dtype)
        for key in ['s2_position_shadow', 's2_time_shadow', 's1_time_shadow']:
            is_position = 'position' in key
            type_str = key.split('_')[0]
            stype = 2 if 's2' in key else 1
            mask_pre = (peaks['type'] == stype) & (peaks['area'] > self.shadow_threshold[key])
            split_peaks = strax.touching_windows(peaks[mask_pre], roi_shadow)
            array = np.zeros(len(current_peak), np.dtype(self.shadowdtype))

            # Initialization
            array['x'] = np.nan
            array['y'] = np.nan
            array['dt'] = self.shadow_time_window_backward
            # The default value for shadow is set to be the lowest possible value
            if 'time' in key:
                array['shadow'] = self.shadow_threshold[key] * array['dt'] ** self.shadow_deltatime_exponent
            else:
                array['shadow'] = 0
            array['nearest_dt'] = self.shadow_time_window_backward

            # Calculating shadow, the Major of the plugin. Only record the previous peak casting the largest shadow
            if len(current_peak):
                self.peaks_shadow(current_peak,
                                  peaks[mask_pre],
                                  split_peaks,
                                  self.shadow_deltatime_exponent,
                                  array,
                                  is_position,
                                  self.getsigma(self.shadow_sigma_and_baseline, current_peak['area']))

            # Fill results
            names = ['shadow', 'dt']
            if 's2' in key:  # Only previous S2 peaks have (x,y)
                names += ['x', 'y']
            if 'time' in key:  # Only time shadow gives the nearest large peak
                names += ['nearest_dt']
            for name in names:
                if name == 'nearest_dt':
                    result[f'{name}_{type_str}'] = array[name]
                else:
                    result[f'{name}_{key}'] = array[name]

        distance = np.sqrt((result[f'x_s2_position_shadow'] - current_peak['x']) ** 2 +
                           (result[f'y_s2_position_shadow'] - current_peak['y']) ** 2)
        # If distance is NaN, set largest distance
        distance = np.where(np.isnan(distance), 2 * straxen.tpc_r, distance)
        # HalfCauchy PDF when calculating S2 position shadow
        result['pdf_s2_position_shadow'] = halfcauchy.pdf(distance,
                                                          scale=self.getsigma(self.shadow_sigma_and_baseline,
                                                                              current_peak['area']))

        # 6. Set time and endtime for peaks
        result['time'] = current_peak['time']
        result['endtime'] = strax.endtime(current_peak)
        return result

    @staticmethod
    @np.errstate(invalid='ignore')
    def getsigma(sigma_and_baseline, s2):
        # The parameter of HalfCauchy, which is a function of S2 area
        return sigma_and_baseline[0] / np.sqrt(s2) + sigma_and_baseline[1]

    @staticmethod
    @numba.njit
    def peaks_shadow(peaks, pre_peaks,
                     touching_windows, exponent,
                     result, pos_corr, sigmas=None):
        """
        For each peak in peaks, check if there is a shadow-casting peak
        and check if it casts the largest shadow
        """
        for p_i, (suspicious_peak, sigma) in enumerate(zip(peaks, sigmas)):
            # casting_peak is the previous large peak casting shadow
            # suspicious_peak is the suspicious peak which in shadow from casting_peak
            indices = touching_windows[p_i]
            for idx in range(indices[0], indices[1]):
                casting_peak = pre_peaks[idx]
                dt = suspicious_peak['center_time'] - casting_peak['center_time']
                if dt <= 0:
                    continue
                # First we record the time difference to the nearest previous peak
                result['nearest_dt'][p_i] = min(result['nearest_dt'][p_i], dt)
                # Calculate time shadow
                new_shadow = casting_peak['area'] * dt ** exponent
                if pos_corr:
                    # Calculate position shadow which is time shadow with a HalfCauchy PDF multiplier
                    distance = distance_in_xy(suspicious_peak, casting_peak)
                    distance = np.where(np.isnan(distance), 2 * straxen.tpc_r, distance)
                    new_shadow *= 2 / (np.pi * sigma * (1 + (distance / sigma) ** 2))
                # Only the previous peak with largest shadow is recorded
                if new_shadow > result['shadow'][p_i]:
                    result['shadow'][p_i] = new_shadow
                    result['x'][p_i] = casting_peak['x']
                    result['y'][p_i] = casting_peak['y']
                    result['dt'][p_i] = suspicious_peak['center_time'] - casting_peak['center_time']



import strax
import straxen
import numpy as np
import numba
import typing as ty
from immutabledict import immutabledict

export, __all__ = strax.exporter()


@export
class nVETOEvents(strax.OverlapWindowPlugin):
    """
    Plugin which computes the boundaries of veto events.
    """
    __version__ = '0.0.3'

    depends_on = 'hitlets_nv'
    provides = 'events_nv'
    data_kind = 'events_nv'
    compressor = 'zstd'
    events_seen = 0

    event_left_extension_nv = straxen.URLConfig(
        default=0,
        track=True,
        type=int,
        help='Extends event window this many [ns] to the left.'
    )

    event_resolving_time_nv = straxen.URLConfig(
        default=200,
        track=True,
        type=int,
        help='Resolving time for window coincidence [ns].'
    )

    event_min_hits_nv = straxen.URLConfig(
        default=3,
        track=True,
        type=int,
        help='Minimum number of fully confined hitlets to define an event.'
    )

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help='immutabledict mapping subdetector to (min, max) channel number'
    )

    def infer_dtype(self):
        self.name_event_number = 'event_number_nv'
        self.channel_range = self.channel_map['nveto']
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_event_dtype(self.name_event_number, self.n_channel)

    def get_window_size(self):
        return self.event_left_extension_nv + self.event_resolving_time_nv + 1

    def compute(self, hitlets_nv, start, end):
        events, hitlets_ids_in_event = find_veto_events(hitlets_nv,
                                                        self.event_min_hits_nv,
                                                        self.event_resolving_time_nv,
                                                        self.event_left_extension_nv,
                                                        event_number_key=self.name_event_number,
                                                        n_channel=self.n_channel, )

        if len(hitlets_ids_in_event):
            compute_nveto_event_properties(events,
                                           hitlets_nv,
                                           hitlets_ids_in_event,
                                           start_channel=self.channel_range[0])

        # Get eventids:
        n_events = len(events)
        events[self.name_event_number] = np.arange(n_events) + self.events_seen
        self.events_seen += n_events

        # Don't extend beyond the chunk boundaries
        # This will often happen for events near the invalid boundary of the
        # overlap processing (which should be thrown away)
        events['time'] = np.clip(events['time'], start, end)
        events['endtime'] = np.clip(events['endtime'], start, end)
        return events


def veto_event_dtype(name_event_number: str = 'event_number_nv',
                     n_pmts: int = 120) -> list:
    dtype = []
    dtype += strax.time_fields  # because mutable
    dtype += [(('Veto event number in this dataset', name_event_number), np.int64),
              (('Total area of all hitlets in event [pe]', 'area'), np.float32),
              (('Total number of hitlets in events', 'n_hits'), np.int32),
              (('Total number of contributing channels', 'n_contributing_pmt'), np.uint8),
              (('Area in event per channel [pe]', 'area_per_channel'), np.float32, n_pmts),
              (('Area weighted mean time of the event relative to the event start [ns]',
                'center_time'), np.float32),
              (('Weighted variance of time [ns]', 'center_time_spread'), np.float32),
              ]
    return dtype


@numba.njit(cache=True, nogil=True)
def compute_nveto_event_properties(events: np.ndarray,
                                   hitlets: np.ndarray,
                                   contained_hitlets_ids: np.ndarray,
                                   start_channel: int = 2000):
    """
    Computes properties of the neutron-veto events. Writes results
    directly to events.

    :param events: Events for which properties should be computed
    :param hitlets: hitlets which were used to build the events.
    :param contained_hitlets_ids: numpy array of the shape n x 2 which holds
        the indices of the hitlets contained in the corresponding event.
    :param start_channel: Integer specifying start channel, e.g. 2000
        for nveto.
    """
    for e, (s_i, e_i) in zip(events, contained_hitlets_ids):
        hitlet = hitlets[s_i:e_i]
        event_area = np.sum(hitlet['area'])
        e['area'] = event_area
        e['n_hits'] = len(hitlet)
        e['n_contributing_pmt'] = len(np.unique(hitlet['channel']))

        t = hitlet['time'] - hitlet[0]['time']
        if event_area:
            e['center_time'] = np.sum(t * hitlet['area']) / event_area
            if e['n_hits'] > 1 and e['center_time']:
                w = hitlet['area'] / e['area']  # normalized weights
                # Definition of variance
                e['center_time_spread'] = np.sqrt(np.sum(w * np.power(t - e['center_time'], 2)) / np.sum(w))
            else:
                e['center_time_spread'] = np.inf

        # Compute per channel properties:
        for hit in hitlet:
            ch = hit['channel'] - start_channel
            e['area_per_channel'][ch] += hit['area']


@export
def find_veto_events(hitlets: np.ndarray,
                     coincidence_level: int,
                     resolving_time: int,
                     left_extension: int,
                     event_number_key: str = 'event_number_nv',
                     n_channel: int = 120, ) -> ty.Tuple[np.ndarray, np.ndarray]:
    """
    Function which find the veto events as a nfold concidence in a given
    resolving time window. All hitlets which touch the event window
    contribute.

    :param hitlets: Hitlets which shall be used for event creation.
    :param coincidence_level: int, coincidence level.
    :param resolving_time: int, resolving window for coincidence in ns.
    :param left_extension: int, left event extension in ns.
    :param event_number_key: str, field name for the event number
    :param n_channel: int, number of channels in detector.
    :returns: events, hitelt_ids_per_event
    """
    # Find intervals which satisfy requirement:
    event_intervals = straxen.plugins.nveto_recorder.find_coincidence(hitlets,
                                                                      coincidence_level,
                                                                      resolving_time,
                                                                      left_extension, )

    # Find all hitlets which touch the coincidence windows:
    # (we cannot use fully_contained in here since some muon signals
    # may be larger than 300 ns)
    hitlets_ids_in_event = strax.touching_windows(hitlets,
                                                  event_intervals)

    # For some rare cases long signals may touch two intervals, in that
    # case we merge the intervals in the subsequent function:
    hitlets_ids_in_event = _solve_ambiguity(hitlets_ids_in_event)

    # Now we can create the veto events:
    events = np.zeros(len(hitlets_ids_in_event),
                      dtype=veto_event_dtype(event_number_key, n_channel))
    _make_event(hitlets, hitlets_ids_in_event, events)
    return events, hitlets_ids_in_event


@numba.njit(cache=True, nogil=False)
def _solve_ambiguity(contained_hitlets_ids: np.ndarray) -> np.ndarray:
    """
    Function which solves the ambiguity if a single hitlets overlaps
    with two event intervals.

    This can happen for muon signals which have a long tail, since we
    define the coincidence window as a fixed window. Hence those tails
    can extend beyond the fixed window.
    """
    res = np.zeros(contained_hitlets_ids.shape, dtype=contained_hitlets_ids.dtype)

    if not len(res):
        # Return empty result
        return res

    offset = 0
    start_i, end_i = contained_hitlets_ids[0]
    for e_i, ids in enumerate(contained_hitlets_ids[1:]):

        if end_i > ids[0]:
            # Current and next interval overlap so just updated the end
            # index.
            end_i = ids[1]
        else:
            # They do not overlap store indices:
            res[offset] = [start_i, end_i]
            offset += 1
            # Init next interval:
            start_i, end_i = ids

    # Last event:
    res[offset, :] = [start_i, end_i]
    offset += 1
    return res[:offset]


@numba.njit(cache=True, nogil=True)
def _make_event(hitlets: np.ndarray,
                hitlet_ids: np.ndarray,
                res: np.ndarray):
    """
    Function which sets veto event time and endtime.
    """
    for ei, ids in enumerate(hitlet_ids):
        hit = hitlets[ids[0]:ids[1]]
        res[ei]['time'] = hit[0]['time']
        res[ei]['endtime'] = np.max(strax.endtime(hit))



from unittest import TestCase, skipIf
from straxen import utilix_is_configured, get_resource
from straxen import InterpolatingMap, save_interpolation_formatted_map


class TestItpMaps(TestCase):
    def open_map(self, map_name, fmt, method='WeightedNearestNeighbors'):
        map_data = get_resource(map_name, fmt=fmt)
        m = InterpolatingMap(map_data, method=method)
        self.assertTrue(m is not None)

    @skipIf(not utilix_is_configured(), 'Cannot download maps without db access')
    def test_lce_map(self):
        self.open_map('XENONnT_s1_xyz_LCE_corrected_qes_MCva43fa9b_wires.json.gz', fmt='json.gz')

    @property
    def _map(self):
        """See https://github.com/XENONnT/straxen/pull/757"""
        map = {
            'coordinate_system': [
                [-18.3, -31.7, -111.5],
                [36.6, -0.0, -111.5],
                [-18.3, 31.7, -111.5],
                [-18.3, -31.7, -37.5],
                [36.6, -0.0, -37.5],
                [-18.3, 31.7, -37.5]],
            'description': 'Array_valued dummy map with lists',
            'name': 'Dummy map',
            'map': [
                [1.7, 11.1],
                [1.7, 11.1],
                [1.7, 11.0],
                [3.3, 5.7],
                [3.3, 5.8],
                [3.3, 5.7],
            ],
        }
        return map

    def test_array_valued(self):
        filename = 'test_array_valued.pkl'
        save_interpolation_formatted_map(
            self._map['map'],
            self._map['coordinate_system'],
            filename=filename,
        )
        filename_quantized = 'test_array_valued_quantized.pkl'
        save_interpolation_formatted_map(
            self._map['map'],
            self._map['coordinate_system'],
            filename=filename_quantized,
            quantum=0.001,
        )
        filename_quantized_gzip = 'test_array_valued_quantized.pkl.gz'
        save_interpolation_formatted_map(
            self._map['map'],
            self._map['coordinate_system'],
            filename=filename_quantized_gzip,
            quantum=0.001,
        )

        # Let's do something easy, check if one fixed point yields the
        # same result if not, our interpolation map depends on the
        # straxen version?! That's bad!
        itp_maps = [
            InterpolatingMap(self._map),
            InterpolatingMap(get_resource(filename, fmt='pkl')),
            InterpolatingMap(get_resource(filename_quantized, fmt='pkl')),
            InterpolatingMap(get_resource(filename_quantized_gzip, fmt='pkl.gz')),
        ]
        for itp_map in itp_maps:
            map_at_random_point = itp_map([[0, 0, 0], [0, 0, -140]])
            places = 5
            self.assertAlmostEqual(map_at_random_point[0][0], 2.80609655, places=places)
            self.assertAlmostEqual(map_at_random_point[0][1], 7.37967879, places=places)

            self.assertAlmostEqual(map_at_random_point[1][0], 2.17815179, places=places)
            self.assertAlmostEqual(map_at_random_point[1][1], 9.47282782, places=places)



import pandas as pd
import numpy as np
import straxen
from immutabledict import immutabledict
import holoviews as hv
import strax

class PlotPMTArrayTPC:
    """
    Class to plot TPC PMT array.
    
    :param gains: If a list of gain values are supplied PMTs which gain
        are set to zero are displayed in gray.
    :param top_pmt_array: Boolean whether top or bottom PMT array should
        be plotted.
    """
    def __init__(self, 
                 gains=None,
                 top_pmt_array=True,
                 opts=immutabledict(),
                ):
    
        self.is_top_pmt_array = top_pmt_array
        
        self._init_start_end_index()
        self._init_pmt_array(gains)
        self._init_common_options(opts)
        
    def _init_start_end_index(self):
        """Defines index-range of channels to be used for the 
        plot. Depends on whether the top or bottom array should be 
        plotted.
        """
        self.pmt_start = 0 if self.is_top_pmt_array else straxen.n_top_pmts+1
        self.pmt_end = straxen.n_top_pmts if self.is_top_pmt_array else straxen.n_tpc_pmts
        
        
    def _init_pmt_array(self, gains):
        """Initalized a dataframe which will contain the Top PMT 
        pattern array.
        """
        _pmts = pd.DataFrame()
        _pmts['channel'] = np.arange(0, straxen.n_tpc_pmts, 1)
        _pmts['x'] = straxen.pmt_positions().loc[:, 'x'].values
        _pmts['y'] = straxen.pmt_positions().loc[:, 'y'].values
        _pmts['gains'] = np.ones(straxen.n_tpc_pmts)
        if np.any(gains):
            if len(gains) != straxen.n_tpc_pmts:
                raise ValueError('Expected gains to have the same shape as number '
                                 'of PMTs in TPC!')
            _is_zero = gains == 0
            _pmts.loc[_is_zero, 'gains'] = np.nan
        self.pmts = _pmts.loc[self.pmt_start:self.pmt_end-1, :]
        
    def _init_common_options(self, opt_dict):
        """Creates dictionary with common options required by PMT array 
        plot.
        """
        if not isinstance(opt_dict, (dict, immutabledict)):
            raise ValueError('opts_dict must be a dictionary!')        
        
        if self.is_top_pmt_array:
            title = 'Top array top view'
        else:
            title = 'Bottom array top view'
            
        r = straxen.tpc_r
        r*=1.1
        default_settings = {'xlim': (-r, r),
                            'ylim': (-r, r),
                            'xlabel': 'x [cm]',
                            'ylabel': 'y [cm]',
                            'title': title
                           }
        self.settings = immutabledict({**default_settings, 
                                       **opt_dict})

    @staticmethod
    def _plot_tpc(**opts):
        """Function which plots the TPC as a simple circle.
        """
        tpc = straxen.holoviews_utils.plot_tpc_circle(straxen.tpc_r)
        tpc.opts(**opts)
        return tpc
    
    def _plot_top_array(self, peak_data, label, **opts):
        pmts = pd.DataFrame()
        pmts['area'] = peak_data['area_per_channel'][self.pmt_start:self.pmt_end]
        
        # Cops other static values: 
        # (we cannot use self.pmts as plot input as new plots will also change prev. 
        # plots due to same reference.)
        pmts['channel'] = self.pmts['channel']
        pmts['x'] = self.pmts['x']
        pmts['y'] = self.pmts['y']
        pmts['gains'] = self.pmts['gains']
        
        
        pmts['area'] /= pmts['gains'] 
        pmt_array = Circle(pmts, 
                           kdims = [hv.Dimension('x',label='x [cm]')],
                           vdims = [hv.Dimension('y', label='y [cm]'),
                                   hv.Dimension('area', label='Area [pe]'),
                                   hv.Dimension('channel', label='Channel')],
                           label=label,
                          )
        pmt_array.opts(color='area',
                       cmap='viridis',
                       line_color='black',
                       radius=3*2.54/2,
                       tools=['hover'],
                       clim=(10, None),
                       clipping_colors={'NaN': 'gray'},
                      )
        pmt_array.opts(**opts)
        return pmt_array
    
    def plot_pmt_array(self, peak, label='', **opts):
        """Function which plots the specified PMT array for the given 
        peak.
        
        :param peak: Peak for which PMT array should be plotted. Must 
            contain area per channel information.
        :param label: Plot abel to be used as title.
        :param opts: Option which can be supplied to a holovies plot.
            E.g. you can use logz=True for a logarithmic colorscale or
            colorbar=True for adding a colorbar.
        :returns: Holoviews plot overlay.
        """
        # Allow to return empty plots required for event display if alt.
        if peak is None:
            return hv.Points(None)
        
        plot = self._plot_tpc() 
        plot *= self._plot_top_array(peak, label=label).opts(**self.settings, **opts)
        return plot
    
# Define function which plots points with radius defined in data space:
# Taken from: https://stackoverflow.com/questions/60361810/how-do-i-set-the-scatter-circle-radius-in-holoviews
import param
from holoviews.element.chart import Chart
from holoviews.plotting.bokeh import PointPlot

class Circle(Chart):
    group = param.String(default='Circle', constant=True)
    size = param.Integer()


class CirclePlot(PointPlot):
    _plot_methods = dict(single='circle', batched='circle')

    style_opts = ['radius' if so == 'size' else so for so in PointPlot.style_opts if so != 'marker']
hv.Store.register({Circle: CirclePlot}, 'bokeh')



import strax
import straxen

import numpy as np
from straxen.plugins.defaults import DEFAULT_POSREC_ALGO

export, __all__ = strax.exporter()


@export
class PeakPositionsNT(strax.MergeOnlyPlugin):
    """
    Merge the reconstructed algorithms of the different algorithms
    into a single one that can be used in Event Basics.

    Select one of the plugins to provide the 'x' and 'y' to be used
    further down the chain. Since we already have the information
    needed here, there is no need to wait until events to make the
    decision.

    Since the computation is trivial as it only combined the three
    input plugins, don't save this plugins output.
    """
    provides = "peak_positions"
    depends_on = ("peak_positions_cnn", "peak_positions_mlp", "peak_positions_gcn")
    save_when = strax.SaveWhen.NEVER
    __version__ = '0.0.0'

    default_reconstruction_algorithm = straxen.URLConfig(
        default=DEFAULT_POSREC_ALGO,
        help="default reconstruction algorithm that provides (x,y)"
    )

    def infer_dtype(self):
        dtype = strax.merged_dtype([self.deps[d].dtype_for(d) for d in self.depends_on])
        dtype += [('x', np.float32, 'Reconstructed S2 X position (cm), uncorrected'),
                  ('y', np.float32, 'Reconstructed S2 Y position (cm), uncorrected')]
        return dtype

    def compute(self, peaks):
        result = {dtype: peaks[dtype] for dtype in peaks.dtype.names}
        algorithm = self.default_reconstruction_algorithm
        for xy in ('x', 'y'):
            result[xy] = peaks[f'{xy}_{algorithm}']
        return result



import strax
import straxen
from straxen import pre_apply_function

export, __all__ = strax.exporter()


@export
class EventInfo(strax.MergeOnlyPlugin):
    """
    Plugin which merges the information of all event data_kinds into a
    single data_type.
    """
    depends_on = ['event_basics',
                  'event_positions',
                  'corrected_areas',
                  'energy_estimates',
                  # 'event_pattern_fit', <- this will be added soon
                  ]
    save_when = strax.SaveWhen.ALWAYS
    provides = 'event_info'
    __version__ = '0.0.2'

    event_info_function = straxen.URLConfig(
        default='pre_apply_function',
        infer_type=False,
        help="Function that must be applied to all event_info data. Do not change.",
    )

    def compute(self, **kwargs):
        event_info_function = self.config['event_info_function']
        event_info = super().compute(**kwargs)
        if event_info_function != 'disabled':
            event_info = pre_apply_function(event_info,
                                            self.run_id,
                                            self.provides,
                                            event_info_function,
                                            )
        return event_info



import numpy as np
import straxen
import strax
import numba

export, __all__ = strax.exporter()


@export
class VetoProximity(strax.OverlapWindowPlugin):
    """
    Find the closest next/previous veto start w.r.t. the event time or
    when a busy happens during an event.
    """

    __version__ = '2.2.0'
    # Strictly speaking, we could depend on 'events', but then you couldn't
    # change the event_window_fields to e.g. s1_time and s2_endtime.
    depends_on = ('event_basics', 'veto_intervals')
    provides = 'veto_proximity'
    data_kind = 'events'

    event_window_fields = straxen.URLConfig(
        default=('time', 'endtime'),
        help='Fields to determine where to look for overlaps for using '
             'this plugin in the events. The default uses start and endtime '
             'of an event, but this can also be the S1 or S2 start/endtime'
    )

    veto_proximity_window = straxen.URLConfig(
        default=int(300e9),
        help='Maximum separation between veto stop and start pulses [ns]'
    )
    time_no_aqmon_veto_found = straxen.URLConfig(
        default=int(3.6e+12),
        track=True,
        type=int,
        help='If no next/previous veto is found, we will fill the fields '
             'time_to_previous_XX with this time. Set to a large number '
             'such that one will never cut events that are < YY ns.'
    )

    veto_names = ['busy', 'busy_he', 'hev', 'straxen_deadtime']

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields
        start_field, stop_field = self.event_window_fields
        for name in self.veto_names:
            dtype += [
                ((f'Duration of event overlapping with "{name}"-veto [ns]',
                  f'veto_{name}_overlap'),
                 np.int64),
                ((f'Time (absolute value) to previous "{name}"-veto '
                  f'from "{start_field}" of event [ns]',
                  f'time_to_previous_{name}'),
                 np.int64),
                ((f'Time (absolute value) to next "{name}"-veto '
                  f'from "{stop_field}" of event [ns]',
                  f'time_to_next_{name}'),
                 np.int64),
            ]

        return dtype

    def get_window_size(self):
        return self.veto_proximity_window

    def set_result_for_veto(self,
                            result_buffer: np.ndarray,
                            event_window: np.ndarray,
                            veto_intervals: np.ndarray,
                            veto_name: str) -> None:
        """
        Fill the result buffer inplace. Goal is to find vetos with
        <veto_name> that are either during, before or after the
         current event_window.

        :param result_buffer: The buffer to fill inplace
        :param event_window: start/stop boundaries of the event to consider.
            Should be an array with ['time'] and ['endtime'] which can be
            based on event start/end times or S1/S2 times
        :param veto_intervals: veto intervals datatype
        :param veto_name: The name of the veto to fill the result buffer for
        :return: Nothing, results are filled in place
        """
        # Set defaults to be some very long time
        result_buffer[f'time_to_previous_{veto_name}'] = self.time_no_aqmon_veto_found
        result_buffer[f'time_to_next_{veto_name}'] = self.time_no_aqmon_veto_found

        selected_intervals = veto_intervals[veto_intervals['veto_type'] == f'{veto_name}_veto']
        if not len(selected_intervals):
            return
        
        vetos_during_event = strax.touching_windows(selected_intervals,
                                                    event_window)
        
        # Figure out the vetos *during* an event
        res = self.get_overlapping_window_time(vetos_during_event, selected_intervals, event_window, result_buffer)
        result_buffer[f'veto_{veto_name}_overlap'] = res

        # Find the next and previous veto's
        times_to_prev, times_to_next = strax.abs_time_to_prev_next_interval(event_window, selected_intervals)
        mask_prev = times_to_prev > 0
        result_buffer[f'time_to_previous_{veto_name}'][mask_prev] = times_to_prev[mask_prev]

        max_next = times_to_next > 0
        result_buffer[f'time_to_next_{veto_name}'][max_next] = times_to_next[max_next]

    @staticmethod
    @numba.njit
    def get_overlapping_window_time(vetos_during_event, selected_intervals, event_window, result_buffer):
        """Computes total time each event overlaps with the corresponding veto.
        """
        res = np.zeros(len(vetos_during_event), np.int64)
        
        for event_i, veto_window in enumerate(vetos_during_event):
            if veto_window[1] - veto_window[0]:
                vetos_in_window = selected_intervals[veto_window[0]:
                                                     veto_window[1]].copy()
                starts = np.clip(vetos_in_window['time'],
                                 event_window[event_i]['time'],
                                 event_window[event_i]['endtime'])
                stops = np.clip(vetos_in_window['endtime'],
                                event_window[event_i]['time'],
                                event_window[event_i]['endtime'])
                
                res[event_i] = np.sum(stops - starts)
        return res

    def compute(self, events, veto_intervals):
        result = np.zeros(len(events), self.dtype)
        result['time'] = events['time']
        result['endtime'] = events['endtime']

        # Get containers for touching windows based on self.event_window_fields
        event_window = np.zeros(len(events), dtype=strax.time_fields)
        event_window['time'] = events[self.event_window_fields[0]]
        event_window['endtime'] = events[self.event_window_fields[1]]

        for veto_name in self.veto_names:
            self.set_result_for_veto(result, event_window, veto_intervals, veto_name)
        return result



import os
import re
import gzip
import json
import time
import pickle
import logging
import warnings
from typing import Type, List, Literal
from textwrap import dedent

import numpy as np
from scipy.spatial import cKDTree
from scipy.interpolate import RectBivariateSpline, RegularGridInterpolator

import straxen
import strax

export, __all__ = strax.exporter()


@export
class InterpolateAndExtrapolate:
    """Linearly interpolate- and extrapolate using inverse-distance
    weighted averaging between nearby points.
    """

    def __init__(self, points, values, neighbours_to_use=None, array_valued=False):
        """
        :param points: array (n_points, n_dims) of coordinates
        :param values: array (n_points) of values
        :param neighbours_to_use: Number of neighbouring points to use for
        averaging. Default is 2 * dimensions of points.
        """
        self.kdtree = cKDTree(points)
        self.values = values
        if neighbours_to_use is None:
            neighbours_to_use = points.shape[1] * 2
        self.neighbours_to_use = neighbours_to_use
        self.array_valued = array_valued
        if array_valued:
            self.n_dim = self.values.shape[-1]

    def __call__(self, points):
        distances, indices = self.kdtree.query(points, self.neighbours_to_use)

        result = np.ones(len(points)) * float('nan')
        if self.array_valued:
            result = np.repeat(result.reshape(-1, 1), self.n_dim, axis=1)

        # If one of the coordinates is NaN, the neighbour-query fails.
        # If we don't filter these out, it would result in an IndexError
        # as the kdtree returns an invalid index if it can't find neighbours.
        valid = (distances < float('inf')).max(axis=-1)

        values = self.values[indices[valid]]
        weights = 1 / np.clip(distances[valid], 1e-6, float('inf'))
        if self.array_valued:
            weights = np.repeat(weights, self.n_dim).reshape(values.shape)

        result[valid] = np.average(
            values, weights=weights, axis=-2 if self.array_valued else -1)
        return result


@export
class InterpolatingMap:
    """Correction map that computes values using inverse-weighted distance
    interpolation.

    The map must be specified as a json translating to a dictionary like this:
        'coordinate_system' :   [[x1, y1], [x2, y2], [x3, y3], [x4, y4], ...],
        'map' :                 [value1, value2, value3, value4, ...]
        'another_map' :         idem
        'name':                 'Nice file with maps',
        'description':          'Say what the maps are, who you are, etc',
        'timestamp':            unix epoch seconds timestamp

    with the straightforward generalization to 1d and 3d.

    Alternatively, a grid coordinate system can be specified as follows:
        'coordinate_system' :   [['x', [x_min, x_max, n_x]], [['y', [y_min, y_max, n_y]]

    Alternatively, an N-vector-valued map can be specified by an array with
    last dimension N in 'map'.

    The default map name is 'map', I'd recommend you use that.

    For a 0d placeholder map, use
        'points': [],
        'map': 42,
        etc

    Default method return inverse-distance weighted average of nearby 2 * dim points
    Extra support includes RectBivariateSpline, RegularGridInterpolator in scipy
    by pass keyword argument like
        method='RectBivariateSpline'

    The interpolators are called with
        'positions' :  [[x1, y1], [x2, y2], [x3, y3], [x4, y4], ...]
        'map_name'  :  key to switch to map interpolator other than the default 'map'
    """
    metadata_field_names = ['timestamp', 'description', 'coordinate_system',
                            'name', 'irregular', 'compressed', 'quantized']

    def __init__(self, data, method='WeightedNearestNeighbors', **kwargs):
        if isinstance(data, bytes):
            data = gzip.decompress(data).decode()
        if isinstance(data, (str, bytes)):
            data = json.loads(data)
        assert isinstance(data, dict), f"Expected dictionary data, got {type(data)}"
        self.data = data

        # Decompress / dequantize the map
        # We should support multiple map names!
        if 'compressed' in self.data:
            compressor, dtype, shape = self.data['compressed']
            self.data['map'] = np.frombuffer(
                strax.io.COMPRESSORS[compressor]['decompress'](self.data['map']),
                dtype=dtype).reshape(*shape)
            del self.data['compressed']
        if 'quantized' in self.data:
            self.data['map'] = self.data['quantized'] * self.data['map'].astype(np.float32)
            del self.data['quantized']

        csys = self.data['coordinate_system']
        if not len(csys):
            self.dimensions = 0
        elif isinstance(csys[0][0], str):
            # Support for specifying coordinate system as a gridspec
            grid = [np.linspace(left, right, points)
                    for _, (left, right, points) in csys]
            csys = np.array(np.meshgrid(*grid, indexing='ij'))
            axes = np.roll(np.arange(len(grid) + 1), -1)
            csys = np.transpose(csys, axes)
            csys = np.array(csys).reshape((-1, len(grid)))
            self.dimensions = len(grid)
        else:
            csys = np.array(csys)
            self.dimensions = len(csys[0])

        self.coordinate_system = csys
        self.interpolators = {}
        self.map_names = sorted([
            k for k in self.data.keys()
            if k not in self.metadata_field_names])

        log = logging.getLogger('InterpolatingMap')
        log.debug(
            'Map name: %s' % self.data.get('name', "NO NAME?!"))
        log.debug(
            'Map description:\n    ' +
            re.sub(r'\n', r'\n    ', self.data.get('description', "NO DESCRIPTION?!")))
        log.debug("Map names found: %s" % self.map_names)

        for map_name in self.map_names:
            # Specify dtype float to set Nones to nan
            map_data = np.array(self.data[map_name], dtype=np.float64)

            if self.dimensions != 0:
                if len(self.coordinate_system) == len(map_data):
                    array_valued = len(map_data.shape) == 2
                else:
                    array_valued = len(map_data.shape) == self.dimensions + 1

            if self.dimensions == 0:
                # 0 D -- placeholder maps which take no arguments
                # and always return a single value
                def itp_fun(positions):
                    return np.array([map_data])

            elif method == 'RectBivariateSpline':
                itp_fun = self._rect_bivariate_spline(csys, map_data, array_valued, **kwargs)

            elif method == 'RegularGridInterpolator':
                itp_fun = self._regular_grid_interpolator(csys, map_data, array_valued, **kwargs)

            elif method == 'WeightedNearestNeighbors':
                itp_fun = self._weighted_nearest_neighbors(csys, map_data, array_valued, **kwargs)

            else:
                raise ValueError(f'Interpolation method {method} is not supported')

            self.interpolators[map_name] = itp_fun

    def __call__(self, *args, map_name='map'):
        """Returns the value of the map at the position given by coordinates
        :param positions: array (n_dim) or (n_points, n_dim) of positions
        :param map_name: Name of the map to use. Default is 'map'.
        """
        return self.interpolators[map_name](*args)

    @staticmethod
    def _rect_bivariate_spline(csys, map_data, array_valued, **kwargs):
        dimensions = len(csys[0])
        grid = [np.unique(csys[:, i]) for i in range(dimensions)]
        grid_shape = [len(g) for g in grid]

        assert dimensions == 2, 'RectBivariateSpline interpolate maps of dimension 2'
        assert not array_valued, 'RectBivariateSpline does not support interpolating array values'
        map_data = map_data.reshape(*grid_shape)
        kwargs = straxen.filter_kwargs(RectBivariateSpline, kwargs)
        rbs = RectBivariateSpline(grid[0], grid[1], map_data, **kwargs)

        def arg_formated_rbs(positions):
            if isinstance(positions, list):
                positions = np.array(positions)
            return rbs.ev(positions[:, 0], positions[:, 1])

        return arg_formated_rbs

    @staticmethod
    def _regular_grid_interpolator(csys, map_data, array_valued, **kwargs):
        dimensions = len(csys[0])
        grid = [np.unique(csys[:, i]) for i in range(dimensions)]
        grid_shape = [len(g) for g in grid]

        if array_valued:
            map_data = map_data.reshape((*grid_shape, map_data.shape[-1]))
        else:
            map_data = map_data.reshape(*grid_shape)

        config = dict(bounds_error=False, fill_value=None)
        kwargs = straxen.filter_kwargs(RegularGridInterpolator, kwargs)
        config.update(kwargs)
        
        return RegularGridInterpolator(tuple(grid), map_data, **config)

    @staticmethod
    def _weighted_nearest_neighbors(csys, map_data, array_valued, **kwargs):
        if array_valued:
            map_data = map_data.reshape((-1, map_data.shape[-1]))
        else:
            map_data = map_data.flatten()
        kwargs = straxen.filter_kwargs(InterpolateAndExtrapolate, kwargs)
        return InterpolateAndExtrapolate(csys, map_data, array_valued=array_valued, **kwargs)

    def scale_coordinates(self, scaling_factor, map_name='map'):
        """Scales the coordinate system by the specified factor
        :params scaling_factor: array (n_dim) of scaling factors if different or single scalar.
        """
        if self.dimensions == 0:
            return
        if hasattr(scaling_factor, '__len__'):
            assert (len(scaling_factor) == self.dimensions), \
                f"Scaling factor array dimension {len(scaling_factor)} " \
                f"does not match grid dimension {self.dimensions}"
            self._sf = scaling_factor
        if isinstance(scaling_factor, (int, float)):
            self._sf = [scaling_factor] * self.dimensions

        alt_csys = self.coordinate_system
        for i, gp in enumerate(self.coordinate_system):
            alt_csys[i] = [gc * k for (gc, k) in zip(gp, self._sf)]

        map_data = np.array(self.data[map_name])
        if len(self.coordinate_system) == len(map_data):
            array_valued = len(map_data.shape) == 2
        else:
            array_valued = len(map_data.shape) == self.dimensions + 1
        if array_valued:
            map_data = map_data.reshape((-1, map_data.shape[-1]))
        itp_fun = InterpolateAndExtrapolate(
            points=np.array(alt_csys),
            values=np.array(map_data),
            array_valued=array_valued)
        self.interpolators[map_name] = itp_fun


@export
def save_interpolation_formatted_map(
        itp_map,
        coordinate_system: List,
        filename: str,
        map_name: str = None,
        quantum: float = None,
        quantum_dtype = np.int16,
        map_description: str = '',
        compressor: Literal['bz2', 'zstd', 'blosc', 'lz4'] = 'zstd',
    ):
    """
    Make a straxen-style InterpolatingMap.
    To fit the large XENONnT per-PMT maps into strax_auxiliary files,
    quantized them to values of 1e-5,
    and store the maps as 16-bit integer multiples of 1e-5, instead of 64-bit floats.
    :param itp_map: numpy itp_map or list of floats,
    should follow the shape indicated by coordinate_system
    :param coordinate_system: coordinate system of the itp_map,
    list of [['x', [x_min, x_max, n_x]], [['y', [y_min, y_max, n_y], ...] for each dimension,
    or [[x1, y1], [x2, y2], [x3, y3], [x4, y4], ...]
    :param filename: filename with '.pkl' extension
    :param map_name: name of map
    :param quantum: quantum of the map if quantized
    :param map_description: map's description
    :param compressor: key of compressor in strax.io.COMPRESSORS
    """
    if isinstance(itp_map, list):
        itp_map = np.array(itp_map)

    if isinstance(coordinate_system[0][0], str):
        itp_map_shape = list(itp_map.shape)
        coordinate_shape = [c[1][2] for c in coordinate_system]
        mask = (len(itp_map_shape) == len(coordinate_shape))
        mask &= all([hs == cs for hs, cs in zip(itp_map_shape, coordinate_shape)])
    else:
        itp_map_shape = len(itp_map)
        coordinate_shape = len(coordinate_system)
        mask = (itp_map_shape == coordinate_shape)
    if not mask:
        raise ValueError(
            f"The shape of itp_map: {itp_map_shape} and "
            f"coordinate system: {coordinate_shape} do not match")

    if quantum is None:
        # if quantum is not specified, just use float32
        q = 1
        quantum_dtype = np.float32
        map = itp_map.astype(quantum_dtype)
    else:
        q = quantum
        if not np.issubdtype(quantum_dtype, np.integer):
            raise ValueError(
                "If using quantization, quantum_dtype must be an integer type,"
                f" but it is now {quantum_dtype}")
        encode_until = np.iinfo(quantum_dtype).max * q
        if itp_map.max() > encode_until:
            raise ValueError(
                f"Map maximum value is {itp_map.max():.4f},"
                " can encode values until {encode_until:.4f}")
        map = np.round(itp_map / q).astype(quantum_dtype)

    output = dict(
        coordinate_system=coordinate_system,
        map=strax.io.COMPRESSORS[compressor]['compress'](map),
        description=dedent(map_description),
        timestamp=time.time(),
        compressed=(compressor, quantum_dtype, map.shape),
    )
    if quantum is not None:
        output['quantized'] = q
    if map_name is not None:
        output['name'] = map_name

    if 'pkl' not in filename:
        warnings.warn("Better use .pkl or .pkl.gz extension for map files")
    splitext = os.path.splitext(filename)
    if splitext[-1] == '.gz':
        opener = gzip.open
    else:
        opener = open
    with opener(filename, mode='wb') as f:
        pickle.dump(output, f)
    print(f"Wrote new map file {filename}, {os.path.getsize(filename) / 1e6:.2f} MB")



import strax
import straxen
import numpy as np
from straxen.plugins.aqmon_hits.aqmon_hits import AqmonChannels


class DetectorSynchronization(strax.Plugin):
    """
    Plugin which computes the synchronization delay between TPC and
    vetos.

    Reference:
        * xenon:xenonnt:dsg:mveto:sync_monitor
    """
    __version__ = '0.0.3'
    depends_on = ('raw_records_aqmon',
                  'raw_records_aqmon_nv',
                  'raw_records_aux_mv')
    provides = 'detector_time_offsets'
    data_kind = 'detector_time_offsets'

    tpc_internal_delay = straxen.URLConfig(
        default={'0': 4917, '020380': 10137},
        type=dict,
        track=True,
        help='Internal delay between aqmon and regular TPC channels ins [ns]'
    )
    adc_threshold_nim_signal = straxen.URLConfig(
        default=500,
        type=int,
        track=True,
        help='Threshold in [adc] to search for the NIM signal'
    )
    # This value is only valid for SR0:
    epsilon_offset = straxen.URLConfig(
        default=76,
        type=int,
        track=True,
        help='Measured missing offset for nveto in [ns]'
    )
    sync_max_delay = strax.Config(
        default=11e3,
        help='max delay DetectorSynchronization [ns]')
    sync_expected_min_clock_distance = straxen.URLConfig(
        default=9.9e9,
        help='min clock distance DetectorSynchronization [ns]')
    sync_expected_max_clock_distance = straxen.URLConfig(
        default=10.1e9,
        help='max clock distance DetectorSynchronization [ns]')

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields
        dtype += [(('Time offset for nV to synchronize with TPC in [ns]',
                    'time_offset_nv'), np.int64),
                  (('Time offset for mV to synchronize with TPC in [ns]',
                    'time_offset_mv'), np.int64)
                  ]
        return dtype

    def compute(self,
                raw_records_aqmon,
                raw_records_aqmon_nv,
                raw_records_aux_mv,
                start, end):
        rr_tpc = raw_records_aqmon
        rr_nv = raw_records_aqmon_nv
        rr_mv = raw_records_aux_mv

        extra_offset = 0
        _mask_tpc = (rr_tpc['channel'] == AqmonChannels.GPS_SYNC)
        if not np.any(_mask_tpc):
            # For some runs in the beginning no signal has been acquired here.
            # In that case we have to add the internal DAQ delay as an extra offset later.
            _mask_tpc = (rr_tpc['channel'] == AqmonChannels.GPS_SYNC_AM)
            extra_offset = self.get_delay()

        hits_tpc = self.get_nim_edge(rr_tpc[_mask_tpc], self.config['adc_threshold_nim_signal'])
        hits_tpc['time'] += extra_offset

        _mask_mveto = (rr_mv['channel'] == AqmonChannels.GPS_SYNC_MV)
        hits_mv = self.get_nim_edge(rr_mv[_mask_mveto], self.config['adc_threshold_nim_signal'])

        _mask_nveto = rr_nv['channel'] == AqmonChannels.GPS_SYNC_NV
        hits_nv = self.get_nim_edge(rr_nv[_mask_nveto], self.config['adc_threshold_nim_signal'])
        nveto_extra_offset = 0
        if not len(hits_nv):
            # During SR0 sync signal was not recorded properly for the
            # neutron-veto, hence take waveform itself as "hits".
            _mask_nveto &= rr_nv['record_i'] == 0
            nveto_extra_offset = self.config['epsilon_offset']
            hits_nv = rr_nv[_mask_nveto]
        hits_nv['time'] += nveto_extra_offset

        offsets_mv = self.estimate_delay(hits_tpc, hits_mv)
        offsets_nv = self.estimate_delay(hits_tpc, hits_nv)
        assert len(offsets_mv) == len(offsets_nv), 'Unequal number of sync signals!'

        result = np.zeros(len(offsets_mv), dtype=self.dtype)
        result['time'] = hits_tpc['time']
        result['endtime'] = strax.endtime(hits_tpc)
        result['time_offset_nv'] = offsets_nv
        result['time_offset_mv'] = offsets_mv

        return result

    def get_delay(self):
        delay = 0
        for run_id, _delay in self.config['tpc_internal_delay'].items():
            if int(self.run_id) >= int(run_id):
                delay = _delay
        return delay

    @staticmethod
    def get_nim_edge(raw_records, threshold=500):
        records = strax.raw_to_records(raw_records)
        strax.baseline(records)
        hits = strax.find_hits(records, min_amplitude=threshold)
        return hits

    def estimate_delay(self, hits_det0, hits_det1):
        """
        Function to estimate the average offset between two hits.
        """
        err_value = -10000000000

        offsets = []
        prev_time = 0
        for ind in range(len(hits_det0)):
            offset = self.find_offset_nearest(hits_det1['time'], hits_det0['time'][ind])
            if ind:
                # Cannot compute time to prev for first event
                time_to_prev = hits_det0['time'][ind] - prev_time
            else:
                time_to_prev = 10e9

            # Additional check to avoid spurious signals
            _correct_distance_to_prev_lock = time_to_prev >= self.sync_expected_min_clock_distance
            _correct_distance_to_prev_lock = time_to_prev < self.sync_expected_max_clock_distance
            if (abs(offset) < self.sync_max_delay) & _correct_distance_to_prev_lock:
                offsets.append(offset)
                prev_time = hits_det0['time'][ind]
            else:
                # Add err_value in case offset is not valid
                offsets.append(err_value)
                prev_time = hits_det0['time'][ind]

        return np.array(offsets)

    def find_offset_nearest(self, array, value):
        if not len(array):
            return -self.sync_max_delay
        array = np.asarray(array)
        idx = (np.abs(array - value)).argmin()
        return value - array[idx]



from scipy.special import logsumexp

import strax
import straxen

import numpy as np
import numba


export, __all__ = strax.exporter()


@export
class BayesPeakClassification(strax.Plugin):
    """
    Bayes Peak classification
    Returns the ln probability of a each event belonging to the S1 and S2 class.
    Uses conditional probabilities and data parameterization learned from wfsim data.
    More info can be found here xenon:xenonnt:ahiguera:bayespeakclassification

    :param peaks: peaks
    :param waveforms: peaks waveforms in PE/ns
    :param quantiles: quantiles in ns, calculate from a cumulative sum over the waveform,
                      from zero to the total area with normalized cumulative sum to determine the time
    :returns: the ln probability of a each peak belonging to S1 and S2 class
    """

    provides = 'peak_classification_bayes'
    depends_on = ('peaks',)
    __version__ = '0.0.3'
    dtype = (strax.time_fields
             + [('ln_prob_s1', np.float32, 'S1 ln probability')]
             + [('ln_prob_s2', np.float32, 'S2 ln probability')]
             )

    # Descriptor configs
    bayes_config_file = straxen.URLConfig(
        default='resource://cmt://'
                'bayes_model'
                '?version=ONLINE&run_id=plugin.run_id&fmt=npy',
        help='Bayes model, conditional probabilities tables and Bayes discrete bins'
    )
    bayes_n_nodes = straxen.URLConfig(
        default=50,
        help='Number of attributes(features) per waveform and quantile'
    )
    n_bayes_classes = straxen.URLConfig(
        default=2,
        help='Number of label classes S1(1)/S2(2)'
    )

    def setup(self):
        self.class_prior = np.ones(self.n_bayes_classes) / self.n_bayes_classes
        self.bins = self.bayes_config_file['bins']
        self.cpt = self.bayes_config_file['cprob']

    def compute(self, peaks):
        result = np.zeros(len(peaks), dtype=self.dtype)

        waveforms, quantiles = compute_wf_and_quantiles(peaks, self.bayes_n_nodes)

        ln_prob_s1, ln_prob_s2 = compute_inference(self.bins, self.bayes_n_nodes, self.cpt,
                                                   self.n_bayes_classes, self.class_prior,
                                                   waveforms, quantiles)
        result['time'] = peaks['time']
        result['endtime'] = strax.endtime(peaks)
        result['ln_prob_s1'] = ln_prob_s1
        result['ln_prob_s2'] = ln_prob_s2
        return result


def compute_wf_and_quantiles(peaks: np.ndarray, bayes_n_nodes: int):
    """
    Compute waveforms and quantiles for a given number of nodes(attributes)
    :param peaks:
    :param bayes_n_nodes: number of nodes or attributes
    :return: waveforms and quantiles
    """
    data = peaks['data'].copy()
    data[data < 0.0] = 0.0
    dt = peaks['dt']
    return _compute_wf_and_quantiles(data, dt, bayes_n_nodes)


@numba.njit(cache=True)
def _compute_wf_and_quantiles(data, sample_length, bayes_n_nodes: int):
    waveforms = np.zeros((len(data), bayes_n_nodes))
    quantiles = np.zeros((len(data), bayes_n_nodes))

    num_samples = data.shape[1]
    step_size = int(num_samples / bayes_n_nodes)
    steps = np.arange(0, num_samples + 1, step_size)
    inter_points = np.linspace(0., 1. - (1. / bayes_n_nodes), bayes_n_nodes)
    cumsum_steps = np.zeros(bayes_n_nodes + 1, dtype=np.float64)
    frac_of_cumsum = np.zeros(num_samples + 1)
    sample_number_div_dt = np.arange(0, num_samples + 1, 1)
    for i, (samples, dt) in enumerate(zip(data, sample_length)):
        # reset buffers
        frac_of_cumsum[:] = 0
        cumsum_steps[:] = 0

        frac_of_cumsum[1:] = np.cumsum(samples)
        frac_of_cumsum[1:] = frac_of_cumsum[1:] / frac_of_cumsum[-1]

        cumsum_steps[:-1] = np.interp(inter_points, frac_of_cumsum, sample_number_div_dt * dt)
        cumsum_steps[-1] = sample_number_div_dt[-1] * dt
        quantiles[i] = cumsum_steps[1:] - cumsum_steps[:-1]

        for j in range(bayes_n_nodes):
            waveforms[i][j] = np.sum(samples[steps[j]:steps[j + 1]])
        waveforms[i] /= (step_size * dt)

    return waveforms, quantiles


def compute_inference(bins: int,
                      bayes_n_nodes: int,
                      cpt: np.ndarray,
                      n_bayes_classes: int,
                      class_prior: np.ndarray,
                      waveforms: np.ndarray,
                      quantiles: np.ndarray):
    """
    Bin the waveforms and quantiles according to Bayes bins and compute inference
    :param bins: Bayes bins
    :param bayes_n_nodes: number of nodes or attributes
    :param cpt: conditional probability tables
    :param n_bayes_classes: number of classes
    :param class_prior: class_prior
    :param waveforms: waveforms
    :param quantiles: quantiles
    :return: ln probability per class S1/S2
    """
    # Bin the waveforms and quantiles.
    waveform_bin_edges = bins[0, :][bins[0, :] > -1]
    waveform_num_bin_edges = len(waveform_bin_edges)
    quantile_bin_edges = bins[1, :][bins[1, :] > -1]
    quantile_num_bin_edges = len(quantile_bin_edges)
    waveform_values = np.digitize(waveforms, bins=waveform_bin_edges) - 1
    waveform_values = np.clip(waveform_values, 0, waveform_num_bin_edges - 2)

    quantile_values = np.digitize(quantiles, bins=quantile_bin_edges) - 1
    quantile_values = np.clip(quantile_values, 0, quantile_num_bin_edges - 2)

    lnposterior = get_log_posterior(bayes_n_nodes,
                                    waveforms,
                                    cpt,
                                    waveform_num_bin_edges,
                                    quantile_num_bin_edges,
                                    n_bayes_classes,
                                    waveform_values,
                                    quantile_values,
                                    )

    lnposterior_sumsamples = np.sum(lnposterior, axis=1)
    lnposterior_sumsamples = lnposterior_sumsamples + np.log(class_prior)

    # If you need to numbafy, use
    # github.com/XENONnT/straxen/blob/c06a611a5a4289c65b9830fc1e68027149b6a5fd/straxen/plugins/bayes_peak_classification.py#L119  # noqa
    normalization = logsumexp(lnposterior_sumsamples, axis=1)
    ln_prob_s1 = lnposterior_sumsamples[:, 0] - normalization
    ln_prob_s2 = lnposterior_sumsamples[:, 1] - normalization
    return ln_prob_s1, ln_prob_s2


def get_log_posterior(bayes_n_nodes: int,
                      waveforms: np.ndarray,
                      cpt: np.ndarray,
                      waveform_num_bin_edges: int,
                      quantile_num_bin_edges: int,
                      n_bayes_classes: int,
                      waveform_values: np.ndarray,
                      quantile_values: np.ndarray,
                      ) -> np.ndarray:
    """
    # TODO, add a description what we are computing here

    :param bayes_n_nodes: number of nodes or attributes
    :param waveforms: waveforms
    :param cpt: conditional probability tables
    :param waveform_num_bin_edges: number of bins for waveforms
    :param quantile_num_bin_edges:number of bins for quantiles
    :param n_bayes_classes: number of classes
    :param waveform_values: digitized waveforms
    :param quantile_values: digitized quantiles
    :return: log-posterior for waveforms and quantiles. NB! This is not
        normalized
    """
    wf_posterior = _get_log_posterior(
        nodes=bayes_n_nodes,
        n_bins=waveform_num_bin_edges - 1,
        n_classes=n_bayes_classes,
        cpt=cpt[:bayes_n_nodes],
        wf_len=len(waveforms),
        wf_values=waveform_values
    )
    quantile_posterior = _get_log_posterior(
        nodes=bayes_n_nodes,
        n_bins=quantile_num_bin_edges - 1,
        n_classes=n_bayes_classes,
        cpt=cpt[bayes_n_nodes:],
        wf_len=len(waveforms),
        wf_values=quantile_values
    )

    lnposterior = np.zeros((len(waveforms), bayes_n_nodes * 2, n_bayes_classes))
    lnposterior[:, :bayes_n_nodes] = wf_posterior
    lnposterior[:, bayes_n_nodes:] = quantile_posterior
    return lnposterior


@numba.njit
def _get_log_posterior(nodes, n_bins, n_classes, cpt, wf_len, wf_values):
    """Wrapper for extracting the ln posterior inside compute_inference"""
    ln_posterior = np.zeros((wf_len, nodes, n_classes))
    for i in range(nodes):
        distribution = cpt[i, :n_bins, :]
        ln_posterior[:, i, :] = np.log(distribution[wf_values[:, i], :])
    return ln_posterior



import param
import panel as pn
import holoviews as hv
import numpy as np
from immutabledict import immutabledict

class EventStats(param.Parameterized):
    """Class which displays all events in a user defined data space. 
    Marks position of single selected event depending on the passed
    index or index widget.
    
    :param events: Events for which histogram should be plotted.
    :param default_config: Dictionary containing the default settings
        for the plot.
    :param exclude_fields: A list of fields to be excluded from the 
        x/y-selection widget. Can be uselful as event_info has a lot of
        fields. Can be either a list/tuple or a function. If passed as
        function must take as an argument a list with field names and 
        return a list with field names.
    :param _index_slider: Parameter to pass an external integer slider 
        to class. E.g. required in event display.
    
    Warning:
        In case the initally selected value is outside of the specified
        plotting range the plot will break.
    """
    _x_edges = None
    _y_edges = None
    _entries = None
    _x_field_name = None
    _y_field_name = None
    _x_bins = None
    _y_bins = None
    _x_log = False
    _y_log = False
    _z_log = False

    # Static widgets used for setting plot layout.
    x_field = 'cs1'
    x_bins = param.Tuple(default=(200, 0, 50), label='x binning (#bins, min max)')
    x_log = param.Boolean(default=False, label='Logarithmic x binning (np.logspace)')
    y_field = 'cs2'
    y_bins = param.Tuple(default=(200, 1000, 2500), label='y binning (#bins, min max)')
    y_log = param.Boolean(default=False, label='Logarithmic y binning (np.logspace)')
    clim = param.Tuple(default=(1, None), label='Colorbarlimit (min, max)')
    z_log = param.Boolean(default=True, label='Logarithmic colorbar')
    VALID_CONFIGS = ['x_field', 'x_bins', 'x_log', 'y_field', 'y_bins', 'y_log', 'clim', 'z_log']
    
    def __init__(self, 
                 events, 
                 default_config=immutabledict(),
                 exclude_fields=None,
                 _index_slider=None,
                 **params):
        for key, value in default_config.items():
            if not hasattr(self, key):
                raise ValueError(f'"{key}" is not a valid config. '
                                 f'Only the following attributes are recognized: {self.VALID_CONFIGS}.')
            setattr(self, key, value)
        
        self.exclude_fields = exclude_fields
        self.events = events
        self._create_widgets(self.x_field,
                             self.y_field,
                             _index_slider,
                            )
        super().__init__(**params)
        
        
    def interactive_event_stats(self):
        """An interactive plot which displays a histogram of the 
        supplied data. The user can manually change the axis and binning
        via the provided widgets. This plot is meant to be used as a 
        stand alone outside of the event display.
        """
        stream_hist, stream_marker = self._create_streams()
        hist = hv.DynamicMap(self._plot_hist_data, 
                             streams=stream_hist)
        selected_points = hv.DynamicMap(self._plot_selected_event, 
                                        streams=stream_marker)
        return pn.Column((hist*selected_points).opts(width=400), 
                         self._make_widget_panel(include_index_slider=True))

    def _create_streams(self):
        """Creates stream objects for the histogram and marker plot.
        """
        stream_hist = dict(
            x_field=self.x_field_widget, 
            y_field=self.y_field_widget,
            x_bins=self.param.x_bins, 
            y_bins=self.param.y_bins,
            clim=self.param.clim,
            x_log=self.param.x_log,
            y_log=self.param.y_log,
            z_log=self.param.z_log,
        )
        
        stream_marker=dict(
            x_field=self.x_field_widget, 
            y_field=self.y_field_widget,
            index=self.index_slider
        )
        
        return stream_hist, stream_marker
        
        
        
    def _create_widgets(self, 
                        x_field='cs1', 
                        y_field='cs2', 
                        _index_slider=None,
                       ):
        """Creates widgets for the x and y field as well as an slider
        for the event index if not supplied externally.
        """
        fields = list(self.events.dtype.names)
        if self.exclude_fields:
            if hasattr(self.exclude_fields, '__call__'):
                fields = self.exclude_fields(fields)
            elif isinstance(self.exclude_fields, (list, tuple)):
                fields = self._exclude_fields(fields, self.exclude_fields)
            else:
                raise ValueError('"exclude_fields" must be either a list or a function!')
                
        fields.sort()
        self.x_field_widget = pn.widgets.Select(
            value=x_field, 
            options=fields,
            name='X-field')
        
        self.y_field_widget = pn.widgets.Select(
            value=y_field, 
            options=fields,
            name='Y-field')
        
        if not _index_slider:
            self.index_slider = pn.widgets.IntSlider(start=0, 
                                                     end=len(self.events),
                                                     name='Event Index:'
                                                    )
        else:
            self.index_slider = _index_slider
    
    @staticmethod
    def _exclude_fields(fields, excluded_fields):
        """Function which drops certain fields from 
        """
        dtype = []
        for f in fields:
            if not f in excluded_fields:
                dtype.append(f)
        return dtype
        
    
    def _make_widget_panel(self, include_index_slider=False):
        """Put widgets used in the selection plot into a single panel
        which can be displayed. Adds index slider for stand alone plot.
        """
        widget_panel = pn.Row(
            pn.Column(self.x_field_widget,
                      self.param.x_bins, 
                      self.param.x_log,),
            pn.Column(self.y_field_widget, 
                      self.param.y_bins, 
                      self.param.y_log,),
            pn.Column(self.param.clim, 
                      self.param.z_log,),
            scroll=True,
        )
        
        if include_index_slider:
            widget_panel = pn.Column(self.index_slider, widget_panel)
        
        return widget_panel
        
    def _hist_data(self, x_field, y_field, xbins, ybins, x_log, y_log):
        """Function which checks if underlying data has to be rebinned.
        """
        _hist_has_not_changed = True
        _hist_has_not_changed &= x_field == self._x_field_name
        _hist_has_not_changed &= y_field == self._y_field_name
        _hist_has_not_changed &= xbins == self._x_bins
        _hist_has_not_changed &= x_log == self._x_log
        _hist_has_not_changed &= ybins == self._y_bins
        _hist_has_not_changed &= y_log == self._y_log
        

        if not _hist_has_not_changed:                    
            x_binning = self._get_binning(*xbins, x_log)
            y_binning = self._get_binning(*ybins, y_log)

            entries, x_edges, y_edges = np.histogram2d(
                self.events[x_field], 
                self.events[y_field],
                bins=(x_binning, 
                      y_binning,), 
            )

            self._x_edges = x_edges
            self._y_edges = y_edges
            self._entries = entries.T

            self._x_field_name = x_field
            self._y_field_name = y_field
            self._x_bins = xbins
            self._y_bins = ybins
            self._x_log = x_log
            self._y_log = y_log

    @staticmethod
    def _get_binning(n_bins, bin_min, bin_max, log):
        """Function which will return an array of bin edges.
        """
        if log:
            bin_min = max(10**-3, bin_min)
            bins = np.logspace(np.log10(bin_min),
                               np.log10(bin_max),
                               n_bins)
        else:
            bins = np.linspace(bin_min, 
                               bin_max, 
                               n_bins)
        return bins

    
    def _plot_hist_data(self, 
                  x_field: str,
                  y_field: str,
                  x_bins: tuple,
                  y_bins: tuple,
                  clim: tuple=(1, None),
                  x_log: bool=False,
                  y_log: bool=False,
                  z_log: bool=True,
                 ):
        """Creates histogram based on user specified settings.
        """
        self._hist_data(x_field,
                        y_field, 
                        x_bins, 
                        y_bins,
                        x_log,
                        y_log,
                       )
        hist = hv.QuadMesh(
            (self._x_edges, 
             self._y_edges,
             self._entries),
            [x_field, 
             y_field],
        )
        hist = hist.apply.opts(logz=z_log, 
                               logx=x_log,
                               logy=y_log,
                               clim=clim,
                               cmap='viridis', 
                               colorbar=True,
                               framewise=True,
                               clipping_colors={'min': 'white'},
                       )
        return hist
    
    
    def _plot_selected_event(self, 
                             x_field: str, 
                             y_field:str, 
                             index: int=0):
        """Function which highlights event with selected index in plot.
        """
        point_x = self.events[x_field][index]
        point_y = self.events[y_field][index]
        _is_nan = (np.isnan(point_x) or np.isnan(point_y))
        if not _is_nan:
            selected_points = hv.Points(
                (point_x, 
                 point_y),
                [self.x_field_widget.value,
                 self.y_field_widget.value
                ]
            ).opts(color='red', size=7, alpha=1)

            hline = hv.HLine(point_y).opts(color='red', alpha=0.4)
            vline = hv.VLine(point_x).opts(color='red', alpha=0.4)
        else:
            # In case one of the parameters is not defined 
            # plot invisilbe point
            selected_points = hv.Points(
                (0, 
                 0),
                [self.x_field_widget.value,
                 self.y_field_widget.value
                ]
            ).opts(color='k', alpha=0, size=7)

            hline = hv.HLine(0).opts(color='k', alpha=0)
            vline = hv.VLine(0).opts(color='k', alpha=0)
        
        return selected_points*hline*vline
